from typing import Union, Optional
import tensorflow as tf
import tensorflow.keras.backend as K


def mask_sequence(
    x: tf.Tensor,
    mask: tf.Tensor,
    value: Union[str, float] = 0.0,
    axis: Optional[int] = None,
):
    """为序列条件mask的函数
    mask: 形如(batch_size, seq_len)的0-1矩阵；
    value: mask部分要被替换成的值，可以是'-inf'或'inf'；
    axis: 序列所在轴，默认为1；
    Copy from https://github.com/bojone/bert4keras/blob/master/bert4keras/backend.py#L117
    """
    if mask is None:
        return x
    else:
        if value == "-inf":
            value = -1e12
        elif value == "inf":
            value = 1e12
        value = K.cast(value, K.dtype(x))
        if K.dtype(mask) != K.dtype(x):
            mask = K.cast(mask, K.dtype(x))
        if axis is None:
            axis = 1
        elif axis < 0:
            axis = K.ndim(x) + axis
        if axis <= 0:
            raise ValueError("axis must be greater than 0")
        for _ in range(axis - 1):
            mask = K.expand_dims(mask, 1)
        for _ in range(K.ndim(x) - K.ndim(mask)):
            mask = K.expand_dims(mask, K.ndim(mask))
        return x * mask + value * (1 - mask)


from __future__ import annotations
from typing import Optional

import tensorflow as tf
from tensorflow.keras.metrics import Accuracy, BinaryAccuracy

from .utils import logits2pred


@tf.keras.utils.register_keras_serializable(package="sknlp")
class AccuracyWithLogits(Accuracy):
    def __init__(
        self,
        name: str = "accuracy",
        dtype: Optional[tf.DType] = None,
        activation: str = "linear",
    ) -> None:
        super().__init__(name=name, dtype=dtype)
        self.activation = activation

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        y_pred = logits2pred(y_logits, self.activation)
        super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self) -> dict[str, str]:
        return {**super().get_config(), "activation": self.activation}


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BinaryAccuracyWithLogits(BinaryAccuracy):
    def update_state(
        self,
        y_true: tf.Tensor,
        y_pred: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(
            y_true, tf.math.sigmoid(y_pred), sample_weight=sample_weight
        )



from __future__ import annotations
from typing import Sequence, List, Dict, Any, Optional, Union, Tuple

import numpy as np
import pandas as pd
from tabulate import tabulate
import tensorflow as tf
import tensorflow_addons as tfa

from sknlp.data import SimilarityDataset
from sknlp.layers import MLPLayer
from sknlp.metrics import BinaryAccuracyWithLogits
from sknlp.utils.classification import (
    classification_fscore,
    logits2probabilities,
    probabilities2classes,
)

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec


class DeepDiscriminator(SupervisedNLPModel):
    dataset_class = SimilarityDataset

    def __init__(
        self,
        classes: Sequence[str] = ("相似度",),
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        text2vec: Optional[Text2vec] = None,
        loss: Optional[str] = None,
        loss_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(
            list(classes),
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            text2vec=text2vec,
            task="similarity",
            **kwargs,
        )
        self._loss = loss
        self._loss_kwargs = loss_kwargs

    def get_loss(self, *args, **kwargs):
        if self._loss == "focal":
            return tfa.losses.SigmoidFocalCrossEntropy(
                from_logits=True, **self._loss_kwargs
            )
        return tf.keras.losses.BinaryCrossentropy(from_logits=True)

    def get_metrics(self, *args, **kwargs) -> List[tf.keras.metrics.Metric]:
        return [BinaryAccuracyWithLogits()]

    def get_monitor(cls) -> str:
        return "val_binary_accuracy"

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        return MLPLayer(
            self.num_fc_layers,
            hidden_size=self.fc_hidden_size,
            output_size=self.num_classes,
            activation=self.fc_activation,
            name="mlp",
        )(inputs)

    def predict_proba(
        self,
        X: Sequence[Tuple[str, str]] = None,
        *,
        dataset: SimilarityDataset = None,
        batch_size: int = 128,
    ) -> np.ndarray:
        logits = super().predict(X=X, dataset=dataset, batch_size=batch_size)
        return logits2probabilities(logits, True)

    def predict(
        self,
        X: Sequence[Tuple[str, str]] = None,
        *,
        dataset: SimilarityDataset = None,
        thresholds: float = 0.5,
        batch_size: int = 128,
    ) -> List[float]:
        return self.predict_proba(X=X, dataset=dataset, batch_size=batch_size)

    def score(
        self,
        X: Sequence[Tuple[str, str]] = None,
        y: Sequence[float] = None,
        *,
        dataset: SimilarityDataset = None,
        thresholds: float = 0.5,
        batch_size: int = 128,
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        probs = self.predict(dataset=dataset, batch_size=batch_size)
        predictions = probabilities2classes(probs, True, thresholds=thresholds)
        return classification_fscore(dataset.y, predictions, [0, 1])

    @classmethod
    def format_score(cls, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "DeepDiscriminator":
        config.pop("task", None)
        config.pop("algorithm", None)
        return super().from_config(config)


from __future__ import annotations
from typing import Any, Sequence, Optional

import tensorflow as tf
from tensorflow.keras.layers import Dropout

from sknlp.layers import BertPreprocessingLayer
from sknlp.data import BertClassificationDataset
from sknlp.module.text2vec import Bert2vec
from .deep_classifier import DeepClassifier


class BertClassifier(DeepClassifier):
    dataset_class = BertClassificationDataset

    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        max_sequence_length: int = 120,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        fc_activation: str = "tanh",
        dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ) -> None:
        super().__init__(
            classes,
            is_multilabel=is_multilabel,
            algorithm="bert",
            num_fc_layers=num_fc_layers,
            fc_hidden_size=fc_hidden_size,
            fc_activation=fc_activation,
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            **kwargs
        )
        self.dropout = dropout
        self.inputs = tf.keras.Input(shape=(), dtype=tf.string, name="text_input")

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        preprocessing_layer = BertPreprocessingLayer(self.text2vec.vocab.sorted_tokens)
        token_ids = preprocessing_layer(inputs)
        if self.dropout:
            self.text2vec.update_dropout(dropout=self.dropout)
        _, cls = self.text2vec(
            [
                token_ids,
                tf.zeros_like(token_ids, dtype=tf.int64),
            ]
        )
        if self.dropout:
            cls = Dropout(self.dropout, name="cls_dropout")(cls)

        return cls

    def get_config(self) -> dict[str, Any]:
        return {**super().get_config(), "dropout": self.dropout}


from __future__ import annotations
from typing import Sequence, Any, Optional, Union

import numpy as np
import pandas as pd
import tensorflow as tf

from sknlp.data import ClassificationDataset
from sknlp.layers import MLPLayer
from sknlp.losses import MultiLabelCategoricalCrossentropy
from sknlp.metrics import PrecisionWithLogits, RecallWithLogits, FBetaScoreWithLogits
from sknlp.utils.classification import (
    logits2probabilities,
    probabilities2classes,
    classification_fscore,
)

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec


class DeepClassifier(SupervisedNLPModel):
    dataset_class = ClassificationDataset
    dataset_args = ["is_multilabel"]

    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        max_sequence_length: Optional[int] = None,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        fc_activation: str = "tanh",
        text2vec: Optional[Text2vec] = None,
        loss: Optional[str] = None,
        loss_kwargs: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        classes = list(classes)
        if "NULL" != classes[0]:
            if "NULL" in classes:
                classes.remove("NULL")
            classes.insert(0, "NULL")
        super().__init__(
            classes,
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            task="classification",
            **kwargs,
        )
        self._is_multilabel = is_multilabel
        self._loss = loss
        self._loss_kwargs = loss_kwargs
        self.num_fc_layers = num_fc_layers
        self.fc_hidden_size = fc_hidden_size
        self.fc_activation = fc_activation

    @property
    def is_multilabel(self) -> bool:
        return self._is_multilabel

    @property
    def thresholds(self) -> list[float]:
        _thresholds = [0.5 for _ in range(self.num_classes)]
        _kthresholds: Optional[dict[str, float]] = self.prediction_kwargs.get(
            "thresholds", None
        )
        if _kthresholds is not None:
            for c, t in _kthresholds.items():
                _thresholds[self.class2idx(c)] = t
        return _thresholds

    @thresholds.setter
    def thresholds(self, thresholds: list[float]) -> None:
        if len(thresholds) != self.num_classes:
            raise ValueError(f"类别数为{self.num_classes}, 但thresholds长度为{len(thresholds)}")
        self.prediction_kwargs["thresholds"] = dict(zip(self.classes, thresholds))

    def get_loss(self, *args, **kwargs) -> list[tf.keras.losses.Loss]:
        if self.is_multilabel:
            if self._loss == "binary_crossentropy":
                return tf.keras.losses.BinaryCrossentropy(from_logits=True)
            return MultiLabelCategoricalCrossentropy()
        else:
            return tf.keras.losses.CategoricalCrossentropy(from_logits=True)

    def get_metrics(self, *args, **kwargs) -> list[tf.keras.metrics.Metric]:
        activation = "sigmoid" if self.is_multilabel else "softmax"
        class_id = None
        if self.num_classes == 2:
            class_id = 1
        return [
            PrecisionWithLogits(activation=activation, class_id=class_id),
            RecallWithLogits(activation=activation, class_id=class_id),
            FBetaScoreWithLogits(
                self.num_classes, activation=activation, class_id=class_id
            ),
        ]

    def get_monitor(cls) -> str:
        return "val_fbeta_score"

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        return MLPLayer(
            self.num_fc_layers,
            hidden_size=self.fc_hidden_size,
            output_size=self.num_classes,
            activation=self.fc_activation,
            name="mlp",
        )(inputs)

    def predict_proba(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        batch_size: int = 128,
    ) -> np.ndarray:
        logits = super().predict(X=X, dataset=dataset, batch_size=batch_size)
        return logits2probabilities(logits, self.is_multilabel)

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, list[float], None] = None,
        batch_size: int = 128,
    ) -> Union[list[str], list[list[str]]]:
        probabilities = self.predict_proba(X=X, dataset=dataset, batch_size=batch_size)
        predictions = probabilities2classes(
            probabilities, self.is_multilabel, thresholds=thresholds or self.thresholds
        )
        if self.is_multilabel:
            return [
                [self.idx2class(i) for i in prediction] for prediction in predictions
            ]
        else:
            return [self.idx2class(i) for i in predictions if self.idx2class(i)]

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, list[float], None] = None,
        batch_size: int = 128,
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        predictions = self.predict(
            dataset=dataset, thresholds=thresholds, batch_size=batch_size
        )
        return classification_fscore(
            dataset.y,
            predictions,
            self.classes[1:] if self.num_classes > 2 else self.classes,
        )

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "is_multilabel": self.is_multilabel,
        }

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "DeepClassifier":
        config.pop("task", None)
        config.pop("algorithm", None)
        return super().from_config(config)


from __future__ import annotations
from typing import Optional

import tensorflow as tf
from tensorflow.keras.layers import Embedding

from sknlp.vocab import Vocab
from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint

from .text2vec import Text2vec


class Word2vec(Text2vec):
    def __init__(
        self,
        vocab: Vocab,
        embedding_size: int,
        segmenter: str = "jieba",
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        embeddings_initializer: WeightInitializer = "uniform",
        embeddings_regularizer: Optional[WeightRegularizer] = None,
        embeddings_constraint: Optional[WeightConstraint] = None,
        name: str = "word2vec",
        **kwargs,
    ) -> None:
        """
        基础符号->向量模块.

        Parameters:
        ----------
        vocab: sknlp.vocab.Vocab. 符号表.
        embed_size: int > 0. 向量维度.
        name: str. 模块名.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
        the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
        the `embeddings` matrix.
        input_length: Length of input sequences, when it is constant.
        This argument is required if you are going to connect
        `Flatten` then `Dense` layers upstream
        (without it, the shape of the dense outputs cannot be computed).

        Inputs
        ----------
        2D tensor with shape: `(batch_size, input_length)`.

        Outputs
        ----------
        3D tensor with shape: `(batch_size, input_length, embed_size)`.
        """
        super().__init__(
            vocab,
            segmenter=segmenter,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            embedding_size=embedding_size,
            name=name,
            **kwargs,
        )
        self.embedding = Embedding(
            len(vocab),
            embedding_size,
            embeddings_initializer=embeddings_initializer,
            embeddings_regularizer=embeddings_regularizer,
            embeddings_constraint=embeddings_constraint,
            mask_zero=True,
            name="embeddings",
        )
        self._model = tf.keras.Sequential(self.embedding, name=name)

    def compute_mask(
        self, inputs: tf.Tensor, mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        return self._model.get_layer("embeddings").compute_mask(inputs, mask=mask)


from __future__ import annotations
from typing import Optional, Any
import logging
import tensorflow as tf

from sknlp.utils.tagging import (
    convert_ids_to_tags,
    tagging_fscore,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class TaggingFScoreMetric(tf.keras.callbacks.Callback):
    def __init__(self, classes: list[str], add_start_end_tag: bool) -> None:
        super().__init__()
        self.classes = classes
        self.add_start_end_tag = add_start_end_tag

    def on_epoch_end(self, epoch: int, logs: Optional[dict[str, Any]] = None) -> None:
        if self.validation_data is None:
            return

        idx2class = dict(zip(range(len(self.classes)), self.classes))
        tag_ids_list = self.model.predict(self.validation_data)
        predictions = []
        for tag_ids in tag_ids_list:
            predictions.append(
                convert_ids_to_tags(
                    tag_ids.numpy().tolist(),
                    lambda x: idx2class[x],
                    add_start_end_tag=self.add_start_end_tag,
                )
            )

        y = []
        for _, tag_ids_array in self.validation_data.as_numpy_iterator():
            y.extend(
                [
                    convert_ids_to_tags(
                        tag_ids,
                        lambda x: idx2class[x],
                        add_start_end_tag=self.add_start_end_tag,
                    )
                    for tag_ids in tag_ids_array.tolist()
                ]
            )

        score_df = tagging_fscore(y, predictions, self.classes[1:])
        logger.info(score_df)
        row = score_df[score_df["class"] == "avg"]
        for col in score_df.columns:
            if col == "support":
                continue
            logs[f"val_{col}"] = row[col].values.tolist()[0]


from __future__ import annotations
from typing import Callable, Optional, Any
import os
import tempfile
from enum import Enum
from collections import Counter

import tensorflow as tf
from tensorflow.keras.initializers import TruncatedNormal

from official.nlp.bert.configs import BertConfig

from sknlp.vocab import Vocab
from sknlp.activations import gelu
from sknlp.layers import BertEncodeLayer
from sknlp.layers.utils import (
    convert_bert_checkpoint,
    convert_electra_checkpoint,
    convert_albert_checkpoint,
)
from .text2vec import Text2vec


def create_checkpoint_file(checkpoint: str) -> None:
    checkpoint_filename = os.path.join(checkpoint, "checkpoint")
    if os.path.exists(checkpoint_filename):
        return

    for filename in os.listdir(checkpoint):
        if "ckpt" in filename:
            model_checkpoint = filename[: filename.index("ckpt") + 4]
            with open(checkpoint_filename, "w") as f:
                f.write(f'model_checkpoint_path: "{model_checkpoint}"\n')
                f.write(f'all_model_checkpoint_paths: "{model_checkpoint}"')
            return


def get_bert_config_filename(
    checkpoint: str, config_filename: Optional[str] = None
) -> str:
    if config_filename is not None and config_filename.endswith("json"):
        return config_filename

    json_files: list[str] = [
        filename for filename in os.listdir(checkpoint) if filename.endswith("json")
    ]
    if not json_files:
        raise FileNotFoundError(f"{checkpoint}中缺少模型配置文件")
    if len(json_files) == 1:
        return json_files[0]
    else:
        for json_file in json_files:
            if "generator" not in json_file:
                return json_file


class BertFamily(Enum):
    BERT = 1
    ALBERT = 2
    ELECTRA = 3


class Bert2vec(Text2vec):
    def __init__(
        self,
        vocab: Vocab,
        segmenter: Optional[str] = None,
        hidden_size: int = 768,
        embedding_size: Optional[int] = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: Optional[int] = None,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: Callable = gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs: bool = False,
        share_layer: bool = False,
        cls_pooling: bool = True,
        name: str = "bert2vec",
        **kwargs,
    ) -> None:
        if embedding_size is None:
            embedding_size = hidden_size
        super().__init__(
            vocab,
            segmenter=segmenter,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            embedding_size=embedding_size,
            name=name,
            **kwargs,
        )
        self.return_all_encoder_outputs = return_all_encoder_outputs
        self.bert_layer = BertEncodeLayer(
            len(self.vocab.sorted_tokens),
            embedding_size=embedding_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            type_vocab_size=type_vocab_size,
            intermediate_size=intermediate_size,
            activation=activation,
            dropout_rate=dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer=initializer,
            share_layer=share_layer,
            cls_pooling=cls_pooling,
            name=name,
        )
        token_ids = tf.keras.Input(
            shape=(sequence_length,), dtype=tf.int64, name="input_token_ids"
        )
        type_ids = tf.keras.Input(
            shape=(sequence_length,), dtype=tf.int64, name="input_type_ids"
        )
        self.inputs = [token_ids, type_ids]
        self._model = tf.keras.Model(
            inputs=self.inputs, outputs=self.bert_layer(self.inputs), name=name
        )

    def __call__(self, inputs: list[tf.Tensor]) -> list[tf.Tensor]:
        outputs = super().__call__(inputs)
        if not self.return_all_encoder_outputs:
            return [outputs[0][-1], outputs[1]]
        return outputs

    def compute_mask(self, inputs: tf.Tensor, mask: Optional[tf.Tensor]) -> tf.Tensor:
        return self._model.get_layer(self._name).compute_mask(inputs, mask=mask)[0]

    def update_dropout(self, dropout: float) -> None:
        bert_layer = self._model.get_layer(self.name)
        bert_layer.embedding_dropout_layer.rate = dropout
        transformer_layers = []
        layer = getattr(bert_layer, "shared_layer", None)
        if layer:
            transformer_layers.append(layer)
        else:
            transformer_layers.extend(getattr(bert_layer, "transformer_layers", []))
        for layer in transformer_layers:
            layer._attention_dropout.rate = dropout
            layer._output_dropout.rate = dropout

    @classmethod
    def from_tfv1_checkpoint(
        cls,
        model_type: BertFamily,
        v1_checkpoint: str,
        config_filename: Optional[str] = None,
        cls_pooling: bool = True,
        sequence_length: Optional[int] = None,
        name: str = "bert2vec",
    ):
        create_checkpoint_file(v1_checkpoint)
        config_filename = get_bert_config_filename(
            v1_checkpoint, config_filename=config_filename
        )
        config = BertConfig.from_json_file(os.path.join(v1_checkpoint, config_filename))
        with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
            token_list = f.read().strip("\n").split("\n")
            vocab = Vocab(
                Counter(token_list), pad_token=token_list[0], unk_token=token_list[1]
            )
        share_layer, cls_pooling = False, True
        convert_checkpoint = convert_bert_checkpoint
        if model_type is BertFamily.ALBERT:
            share_layer = True
            convert_checkpoint = convert_albert_checkpoint
        if model_type is BertFamily.ELECTRA:
            cls_pooling = False
            convert_checkpoint = convert_electra_checkpoint
        activation = tf.keras.activations.get(config.hidden_act)
        module = cls(
            vocab,
            embedding_size=config.embedding_size,
            hidden_size=config.hidden_size,
            num_layers=config.num_hidden_layers,
            num_attention_heads=config.num_attention_heads,
            intermediate_size=config.intermediate_size,
            activation=activation,
            dropout_rate=config.hidden_dropout_prob,
            attention_dropout_rate=config.attention_probs_dropout_prob,
            sequence_length=sequence_length,
            max_sequence_length=config.max_position_embeddings,
            type_vocab_size=config.type_vocab_size,
            initializer=TruncatedNormal(stddev=config.initializer_range),
            share_layer=share_layer,
            cls_pooling=cls_pooling,
            name=name,
        )
        with tempfile.TemporaryDirectory() as temp_dir:
            temporary_checkpoint = os.path.join(temp_dir, "ckpt")
            convert_checkpoint(
                checkpoint_from_path=v1_checkpoint,
                checkpoint_to_path=temporary_checkpoint,
                num_heads=config.num_attention_heads,
                converted_root_name=name,
            )
            module._model.load_weights(
                temporary_checkpoint
            ).assert_existing_objects_matched()
        return module

    def export(
        self,
        directory: str,
        name: str,
        version: str = "0",
        only_output_cls: bool = False,
    ) -> None:
        original_model = self._model
        if only_output_cls:
            self._model = tf.keras.Model(
                inputs=self._model.inputs,
                outputs=self._model.outputs[1],
                name="bert2vec",
            )
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self.save_vocab(d)
        self._model = original_model

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "return_all_encoder_outputs": self.return_all_encoder_outputs,
        }


from __future__ import annotations
from typing import Any

import tensorflow as tf

from sknlp.layers import LSTMP
from .deep_tagger import DeepTagger


class TextRNNTagger(DeepTagger):
    def __init__(
        self,
        classes: list[str],
        max_sequence_length: int = 100,
        output_format: str = "global_pointer",
        crf_learning_rate_multiplier: float = 1.0,
        num_rnn_layers: int = 1,
        rnn_hidden_size: int = 512,
        rnn_projection_size: int = 128,
        rnn_recurrent_clip: float = 3.0,
        rnn_projection_clip: float = 3.0,
        rnn_kernel_initializer="glorot_uniform",
        rnn_recurrent_initializer="orthogonal",
        rnn_projection_initializer="glorot_uniform",
        rnn_bias_initializer="zeros",
        rnn_kernel_regularizer=None,
        rnn_recurrent_regularizer=None,
        rnn_projection_regularizer=None,
        rnn_bias_regularizer=None,
        rnn_kernel_constraint=None,
        rnn_recurrent_constraint=None,
        rnn_projection_constraint=None,
        rnn_bias_constraint=None,
        dropout: float = 0.5,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 128,
        fc_activation: str = "tanh",
        text2vec=None,
        **kwargs
    ):
        super().__init__(
            classes,
            add_start_end_tag=False,
            output_format=output_format,
            crf_learning_rate_multiplier=crf_learning_rate_multiplier,
            max_sequence_length=max_sequence_length,
            num_fc_layers=num_fc_layers,
            fc_hidden_size=fc_hidden_size,
            fc_activation=fc_activation,
            text2vec=text2vec,
            algorithm="rnn",
            **kwargs
        )
        self.dropout = dropout
        self.num_rnn_layers = num_rnn_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.rnn_projection_size = rnn_projection_size
        self.rnn_recurrent_clip = rnn_recurrent_clip
        self.rnn_projection_clip = rnn_projection_clip
        self.rnn_kernel_initializer = rnn_kernel_initializer
        self.rnn_recurrent_initializer = rnn_recurrent_initializer
        self.rnn_projection_initializer = rnn_projection_initializer
        self.rnn_bias_initializer = rnn_bias_initializer
        self.rnn_kernel_regularizer = rnn_kernel_regularizer
        self.rnn_recurrent_regularizer = rnn_recurrent_regularizer
        self.rnn_projection_regularizer = rnn_projection_regularizer
        self.rnn_bias_regularizer = rnn_bias_regularizer
        self.rnn_kernel_constraint = rnn_kernel_constraint
        self.rnn_recurrent_constraint = rnn_recurrent_constraint
        self.rnn_projection_constraint = rnn_projection_constraint
        self.rnn_bias_constraint = rnn_bias_constraint
        if self.output_format == "bio":
            self.inputs = [
                tf.keras.Input(shape=(None,), dtype=tf.int32, name="token_id"),
                tf.keras.Input(shape=(None,), dtype=tf.int32, name="tag_id"),
            ]
        else:
            self.inputs = tf.keras.Input(shape=(None,), dtype=tf.int32, name="token_id")

    def build_encode_layer(self, inputs):
        if self.output_format == "bio":
            token_ids = inputs[0]
        else:
            token_ids = inputs
        mask = tf.keras.layers.Lambda(
            lambda x: tf.cast(x != 0, tf.int32), name="mask_layer"
        )(token_ids)
        outputs = self.text2vec(token_ids)
        for _ in range(self.num_rnn_layers):
            outputs = tf.keras.layers.Bidirectional(
                LSTMP(
                    self.rnn_hidden_size,
                    projection_size=self.rnn_projection_size,
                    recurrent_clip=self.rnn_recurrent_clip,
                    projection_clip=self.rnn_projection_clip,
                    dropout=self.dropout,
                    recurrent_dropout=self.dropout,
                    kernel_initializer=self.rnn_kernel_initializer,
                    recurrent_initializer=self.rnn_recurrent_initializer,
                    projection_initializer=self.rnn_projection_initializer,
                    bias_initializer=self.rnn_bias_initializer,
                    kernel_regularizer=self.rnn_kernel_regularizer,
                    recurrent_regularizer=self.rnn_recurrent_regularizer,
                    projection_regularizer=self.rnn_projection_regularizer,
                    bias_regularizer=self.rnn_bias_regularizer,
                    kernel_constraint=self.rnn_kernel_constraint,
                    recurrent_constraint=self.rnn_recurrent_constraint,
                    projection_constraint=self.rnn_projection_constraint,
                    bias_constraint=self.rnn_bias_constraint,
                    return_sequences=True,
                )
            )(outputs)
        if self.dropout:
            noise_shape = (None, 1, self.rnn_projection_size * 2)
            outputs = tf.keras.layers.Dropout(
                self.dropout, noise_shape=noise_shape, name="encoder_dropout"
            )(outputs)
        if self.output_format == "bio":
            return outputs, mask, inputs[1]
        else:
            return outputs, mask

    def get_config(self) -> dict[str, Any]:
        return {**super().get_config(), "dropout": self.dropout}

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "TextRNNTagger":
        config.pop("add_start_end_tag", None)
        return super().from_config(config)


from __future__ import annotations
from typing import Sequence, Optional, Tuple, Any, Callable
import json

import numpy as np
import tensorflow as tf

from .nlp_dataset import NLPDataset


def _combine_xy(x, y):
    return (x, y), y


class TaggingDataset(NLPDataset):
    def __init__(
        self,
        tokenizer: Callable[[str], list[int]],
        labels: Sequence[str],
        X: Optional[Sequence[Any]] = None,
        y: Optional[Sequence[Any]] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        add_start_end_tag: bool = False,
        output_format: str = "global_pointer",
        max_length: Optional[int] = None,
        text_dtype: tf.DType = tf.int32,
        label_dtype: tf.DType = tf.int32,
    ):
        self.add_start_end_tag = add_start_end_tag
        self.output_format = output_format
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(
            tokenizer,
            X=X,
            y=y,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            max_length=max_length,
            na_value="",
            column_dtypes=["str", "str"],
            text_dtype=text_dtype,
            label_dtype=label_dtype,
        )

    @property
    def y(self) -> list[list[str]]:
        if self.no_label:
            return []
        return [
            json.loads(data[-1].decode("UTF-8"))
            for data in self._original_dataset.as_numpy_iterator()
        ]

    @property
    def batch_padding_shapes(self) -> list[Tuple]:
        if self.output_format == "bio":
            return ((None,), (None,))
        else:
            return ((None,), (None, None, None))

    def _normalize_y(self, y: Sequence[Any]) -> Sequence[Any]:
        if isinstance(y[0], (list, tuple)):
            return [json.dumps(yi) for yi in y]
        return y

    def _label_transform(self, label: tf.Tensor, length: int) -> np.array:
        label = super()._label_transform(label)
        chunks = json.loads(label)
        length += 2 * self.add_start_end_tag
        if self.output_format == "bio":
            labels = np.zeros(length, dtype=np.int32)
            for chunk_start, chunk_end, chunk_label in chunks:
                if chunk_end >= self.max_length:
                    continue

                chunk_start += self.add_start_end_tag
                chunk_end += self.add_start_end_tag
                labels[chunk_start] = self.label2idx[chunk_label] * 2 - 1
                for i in range(chunk_start + 1, chunk_end + 1):
                    labels[i] = self.label2idx[chunk_label] * 2
        else:
            labels = np.zeros((len(self.label2idx), length, length), dtype=np.int32)
            for chunk_start, chunk_end, chunk_label in chunks:
                if chunk_end > self.max_length:
                    continue

                chunk_start += self.add_start_end_tag
                chunk_end += self.add_start_end_tag
                labels[self.label2idx[chunk_label], chunk_start, chunk_end] = 1
        return labels

    def _transform_func(self, *data) -> list[Any]:
        text = data[0]

        _text = self._text_transform(text)
        if self.no_label:
            if self.output_format == "bio":
                return _text, [0 for _ in range(len(_text))]
            else:
                return _text
        label = data[1]
        return _text, self._label_transform(label, len(_text))

    def _transform_func_out_dtype(self) -> list[tf.DType]:
        return (self.text_dtype, self.label_dtype)

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        return super().batchify(
            batch_size,
            shuffle=shuffle,
            shuffle_buffer_size=shuffle_buffer_size,
            after_batch=_combine_xy if self.output_format == "bio" else None,
        )


from __future__ import annotations
from typing import Sequence, Any, Optional

import pandas as pd
import tensorflow as tf

from sknlp.data import TaggingDataset
from sknlp.callbacks import TaggingFScoreMetric
from sknlp.metrics import PrecisionWithLogits, RecallWithLogits, FBetaScoreWithLogits
from sknlp.layers import CrfDecodeLayer, CrfLossLayer, MLPLayer, GlobalPointerLayer
from sknlp.losses import MultiLabelCategoricalCrossentropy
from sknlp.utils.tagging import (
    Tag,
    tagging_fscore,
    convert_ids_to_tags,
    convert_global_pointer_to_tags,
)

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec


class DeepTagger(SupervisedNLPModel):
    dataset_class = TaggingDataset
    dataset_args = ["output_format", "add_start_end_tag"]

    def __init__(
        self,
        classes: Sequence[str],
        add_start_end_tag: bool = False,
        output_format: str = "global_pointer",
        crf_learning_rate_multiplier: float = 1.0,
        max_sequence_length: Optional[int] = None,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 128,
        fc_activation: str = "tanh",
        text2vec: Optional[Text2vec] = None,
        **kwargs,
    ):
        self._add_start_end_tag = add_start_end_tag
        if output_format not in {"bio", "global_pointer"}:
            raise ValueError(
                f"output_format必须为'bio'或者'global_pointer', 目前为{output_format}"
            )
        self._output_format = output_format
        self._crf_learning_rate_multiplier = crf_learning_rate_multiplier
        classes = list(classes)
        if output_format == "bio" and classes[0] != "O":
            classes.insert(0, "O")
        super().__init__(
            classes,
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            task="tagging",
            **kwargs,
        )
        self.num_fc_layers = num_fc_layers
        self.fc_hidden_size = fc_hidden_size
        self.fc_activation = fc_activation

    @property
    def add_start_end_tag(self) -> bool:
        return self._add_start_end_tag

    @property
    def output_format(self) -> str:
        return self._output_format

    @property
    def crf_learning_rate_multiplier(self) -> float:
        return self._crf_learning_rate_multiplier

    def get_loss(self) -> None:
        if self.output_format == "bio":
            return None
        else:
            return MultiLabelCategoricalCrossentropy(flatten_axis=2)

    def get_callbacks(self, *args, **kwargs) -> list[tf.keras.callbacks.Callback]:
        callbacks = super().get_callbacks(*args, **kwargs)
        if self.output_format == "bio":
            callbacks.append(TaggingFScoreMetric(self.classes, self.add_start_end_tag))
        return callbacks

    def get_metrics(self) -> list[tf.keras.metrics.Metric]:
        if self.output_format == "bio":
            return []
        return [
            PrecisionWithLogits(activation="sigmoid"),
            RecallWithLogits(activation="sigmoid"),
            FBetaScoreWithLogits(self.num_classes, activation="sigmoid"),
        ]

    def get_monitor(self) -> str:
        if self.output_format == "bio":
            return "val_tag_accuracy"
        else:
            return "val_fbeta_score"

    def build_output_layer(self, inputs: list[tf.Tensor]) -> list[tf.Tensor]:
        if self.output_format == "bio":
            embeddings, mask, tag_ids = inputs
            emissions = MLPLayer(
                self.num_fc_layers,
                hidden_size=self.fc_hidden_size,
                output_size=self.num_classes * 2 + 1,
                activation=self.fc_activation,
                name="mlp",
            )(embeddings)
            return CrfLossLayer(
                self.num_classes * 2 + 1,
                max_sequence_length=self.max_sequence_length,
                learning_rate_multiplier=self.crf_learning_rate_multiplier,
            )([emissions, tag_ids], mask)
        else:
            embeddings, mask = inputs
            return GlobalPointerLayer(self.num_classes, 64, self.max_sequence_length)(
                embeddings, mask
            )

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: TaggingDataset = None,
        thresholds: float = 0.5,
        batch_size: int = 128,
    ) -> list[list[Tag]]:
        raw_predictions = super().predict(X, dataset=dataset, batch_size=batch_size)
        predictions: list[list[Tag]] = []
        if self.output_format == "bio":
            for tag_ids in raw_predictions:
                predictions.append(
                    convert_ids_to_tags(
                        tag_ids.numpy().tolist(), self.idx2class, self.add_start_end_tag
                    )
                )
        else:
            for pointer in raw_predictions:
                predictions.append(
                    convert_global_pointer_to_tags(
                        pointer.to_tensor(-1e20).numpy(),
                        thresholds,
                        self.idx2class,
                        self.add_start_end_tag,
                    )
                )
        return predictions

    def score(
        self,
        X: Sequence[str] = None,
        y: Sequence[Sequence[str]] = None,
        *,
        dataset: TaggingDataset = None,
        thresholds: float = 0.5,
        batch_size: int = 128,
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        predictions = self.predict(
            dataset=dataset, batch_size=batch_size, thresholds=thresholds
        )
        classes = self.classes
        if self.output_format == "bio":
            classes = classes[1:]
        return tagging_fscore(
            [[Tag(*l) for l in yi] for yi in dataset.y], predictions, classes
        )

    def export(self, directory: str, name: str, version: str = "0") -> None:
        if self.output_format == "bio":
            mask = self._model.get_layer("mask_layer").output
            emissions = self._model.get_layer("mlp").output
            crf = CrfDecodeLayer(
                self.num_classes * 2 + 1,
                learning_rate_multiplier=self.crf_learning_rate_multiplier,
                max_sequence_length=self.max_sequence_length,
            )
            crf.build(
                [tf.TensorShape([None, None, None]), tf.TensorShape([None, None])]
            )
            crf.set_weights(self._model.get_layer("crf").get_weights())
            original_model = self._model
            self._model = tf.keras.Model(
                inputs=self._model.inputs[0], outputs=crf(emissions, mask)
            )
            super().export(directory, name, version=version)
            self._model = original_model
        else:
            original_model = self._model
            self._model = tf.keras.Model(
                inputs=self._model.inputs,
                outputs=self._model.outputs[0].to_tensor(tf.float32.min),
            )
            super().export(directory, name, version=version)
            self._model = original_model

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "add_start_end_tag": self.add_start_end_tag,
            "output_format": self.output_format,
            "crf_learning_rate_multiplier": self._crf_learning_rate_multiplier,
        }

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "DeepTagger":
        config.pop("algorithm", None)
        config.pop("task", None)
        return super().from_config(config)


import tensorflow as tf


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _str_feature(value):
    return _bytes_feature(value.encode('utf-8'))


def _tensor_feature(value):
    return _bytes_feature(tf.io.serialize_tensor(value))


def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


DATA_TYPE = {
    'bytes': _bytes_feature,
    'float': _float_feature,
    'int': _int64_feature,
    'str': _str_feature,
    'tensor': _tensor_feature
}


def serialize_example(values, dtypes, names=None):
    """
    Creates a tf.Example message ready to be written to a file.
    """
    if names is None:
        names = ['feature%d' % i for i in range(len(dtypes))]
    assert len(values) == len(dtypes) == len(names)
    feature = {
        name: DATA_TYPE[dtype](value)
        for value, dtype, name in zip(values, dtypes, names)
    }
    # Create a Features message using tf.train.Example.
    example_proto = tf.train.Example(
        features=tf.train.Features(feature=feature)
    )
    x = example_proto.SerializeToString()
    return x



import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BertTokenizationLayer(tf.keras.layers.Layer):
    """
    The convention in BERT is:
    (a) For sequence pairs:
    tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
    type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
    (b) For single sequences:
    tokens:   [CLS] the dog is hairy . [SEP]
    type_ids: 0     0   0   0  0     0 0

    Where "type_ids" are used to indicate whether this is the first
    sequence or the second sequence. The embedding vectors for `type=0` and
    `type=1` were learned during pre-training and are added to the wordpiece
    embedding vector (and position vector). This is not *strictly* necessary
    since the [SEP] token unambiguously separates the sequences, but it makes
    it easier for the model to learn the concept of sequences.

    For classification tasks, the first vector (corresponding to [CLS]) is
    used as the "sentence vector". Note that this only makes sense because
    the entire model is fine-tuned.

    The mask has 1 for real tokens and 0 for padding tokens. Only real
    tokens are attended to.
    See more details in https://github.com/tensorflow/models/blob/ad423d065701785587d13d0fe7b566191e7378c6/official/nlp/data/classifier_data_lib.py#L668
    """

    def __init__(
        self,
        cls_token: str = "[CLS]",
        sep_token: str = "[SEP]",
        name: str = "bert_tokenization",
        **kwargs
    ) -> None:
        self.cls_token = cls_token
        self.sep_token = sep_token
        super().__init__(name=name, **kwargs)

    def call(self, inputs):
        tokens = tf.strings.unicode_split(inputs, "UTF-8")
        cls_tokens = tf.reshape(
            tf.tile(tf.constant([self.cls_token], dtype=tf.string), [tokens.nrows()]),
            [tokens.nrows(), 1],
        )
        sep_tokens = tf.reshape(
            tf.tile(tf.constant([self.sep_token], dtype=tf.string), [tokens.nrows()]),
            [tokens.nrows(), 1],
        )
        return tf.concat([cls_tokens, tokens, sep_tokens], axis=1)

    def get_config(self):
        return {
            **super().get_config(),
            "sep_token": self.sep_token,
            "cls_token": self.cls_token,
        }



from typing import Callable, Union

import tensorflow as tf


WeightRegularizer = Union[Callable[[tf.Tensor], tf.Tensor], str]
WeightInitializer = Union[Callable[[tf.Tensor, tf.DType], tf.Tensor], str]
WeightConstraint = Union[Callable[[tf.Tensor], tf.Tensor], str]


__all__ = ["WeightRegularizer", "WeightInitializer", "WeightConstraint"]



from __future__ import annotations
from typing import Any, Sequence, List, Optional, Tuple, Callable

import pandas as pd
import tensorflow as tf

from .nlp_dataset import NLPDataset


def _combine_xy(text, context, label):
    return ((text, context), label)


def _combine_x(text, context):
    return ((text, context),)


class SimilarityDataset(NLPDataset):
    def __init__(
        self,
        tokenizer: Callable[[str], list[int]],
        labels: Sequence[str],
        X: Optional[Sequence[Any]] = None,
        y: Optional[Sequence[Any]] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        max_length: Optional[int] = None,
        text_dtype: tf.DType = tf.int32,
        label_dtype: tf.DType = tf.float32,
    ):
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(
            tokenizer,
            X=X,
            y=y,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            max_length=max_length,
            na_value=0.0,
            column_dtypes=["str", "str", "float32"],
            text_dtype=text_dtype,
            label_dtype=label_dtype,
        )

    @property
    def y(self) -> List[float]:
        if self.no_label:
            return []
        return [float(data[-1]) for data in self._original_dataset.as_numpy_iterator()]

    @property
    def batch_padding_shapes(self) -> List[Tuple]:
        return ((None,), (None,), (None,))[: -1 if self.no_label else None]

    def _label_transform(self, label: tf.Tensor) -> float:
        return label

    def _transform_func(self, *data) -> List[Any]:
        text = data[0]
        context = data[1]
        if self.no_label:
            return (
                self._text_transform(text),
                self._text_transform(context),
            )
        label = data[2]
        return (
            self._text_transform(text),
            self._text_transform(context),
            self._label_transform(label),
        )

    def _transform_func_out_dtype(self) -> List[tf.DType]:
        return (self.text_dtype, self.text_dtype, self.label_dtype)[
            : -1 if self.no_label else None
        ]

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        return super().batchify(
            batch_size,
            shuffle=shuffle,
            shuffle_buffer_size=shuffle_buffer_size,
            after_batch=_combine_xy if not self.no_label else _combine_x,
        )


from __future__ import annotations
from typing import Optional, Any

import tensorflow as tf
from tensorflow.keras.metrics import Precision, Recall
from tensorflow_addons.metrics import FBetaScore

from .utils import logits2pred


@tf.keras.utils.register_keras_serializable(package="sknlp")
class PrecisionWithLogits(Precision):
    def __init__(
        self,
        top_k: Optional[int] = None,
        class_id: Optional[int] = None,
        threshold: float = 0.5,
        name: str = "precision",
        dtype: Optional[tf.DType] = None,
        activation: str = "linear",
    ) -> None:
        super().__init__(
            thresholds=threshold,
            top_k=top_k,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )
        self.activation = activation

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        y_pred = logits2pred(y_logits, self.activation)
        super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self) -> dict[str, Any]:
        configs = super().get_config()
        configs.pop("thresholds", None)
        return {
            **configs,
            "threshold": self.thresholds,
            "activation": self.activation,
        }


@tf.keras.utils.register_keras_serializable(package="sknlp")
class RecallWithLogits(Recall):
    def __init__(
        self,
        top_k: Optional[int] = None,
        class_id: Optional[int] = None,
        threshold: float = 0.5,
        name: str = "recall",
        dtype: Optional[tf.DType] = None,
        activation: str = "linear",
    ) -> None:
        super().__init__(
            thresholds=threshold,
            top_k=top_k,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )
        self.activation = activation

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        y_pred = logits2pred(y_logits, self.activation)
        super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self) -> dict[str, Any]:
        configs = super().get_config()
        configs.pop("thresholds", None)
        return {
            **configs,
            "threshold": self.thresholds,
            "activation": self.activation,
        }


@tf.keras.utils.register_keras_serializable(package="sknlp")
class FBetaScoreWithLogits(FBetaScore):
    def __init__(
        self,
        num_classes: int,
        average: str = "micro",
        beta: float = 1.0,
        class_id: Optional[int] = None,
        threshold: float = 0.5,
        name: str = "fbeta_score",
        dtype: Optional[tf.DType] = None,
        activation: str = "linear",
    ) -> None:
        super().__init__(
            num_classes,
            average=average,
            beta=beta,
            threshold=threshold,
            name=name,
            dtype=dtype,
        )
        self.class_id = class_id
        self.activation = activation

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        y_pred = logits2pred(y_logits, self.activation)
        if self.class_id is not None:
            y_true = y_true[..., self.class_id]
            y_pred = y_pred[..., self.class_id]
        super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def reset_state(self):
        return super().reset_states()

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "activation": self.activation,
            "class_id": self.class_id,
        }


from .mlp import MLPLayer
from .lstmp import LSTMPCell, LSTMP
from .multi_lstmp import MultiLSTMP
from .bert_tokenization import BertTokenizationLayer
from .bert_layer import (
    BertEncodeLayer,
    BertPreprocessingLayer,
    BertCharPreprocessingLayer,
    BertPairPreprocessingLayer,
)
from .crf_layer import CrfLossLayer, CrfDecodeLayer
from .global_pointer import GlobalPointerLayer


__all__ = [
    "MLPLayer",
    "LSTMPCell",
    "LSTMP",
    "MultiLSTMP",
    "BertTokenizationLayer",
    "BertEncodeLayer",
    "BertPreprocessingLayer",
    "BertCharPreprocessingLayer",
    "BertPairPreprocessingLayer",
    "CrfLossLayer",
    "CrfDecodeLayer",
    "GlobalPointerLayer",
]



from __future__ import annotations
from typing import Sequence, Callable, Union
from dataclasses import dataclass
import itertools

import pandas as pd
import numpy as np
from scipy.special import expit

from .classification import _validate_thresholds


@dataclass
class Tag:
    start: int
    end: str
    label: str

    def __hash__(self) -> int:
        return hash((self.label, self.start, self.end))


def convert_ids_to_tags(
    tag_ids: Sequence[int],
    idx2class: Callable[[int], str],
    add_start_end_tag: bool = False,
) -> list[Tag]:
    if add_start_end_tag:
        tag_ids = tag_ids[1:-1]
    num_tag_ids = len(tag_ids)
    current_begin_tag = -1
    begin = 0
    parsed_tags = list()
    for i, tag_id in enumerate(tag_ids):
        if (
            i < num_tag_ids - 1
            and tag_id != 0
            and tag_id % 2 == 0
            and tag_id - 1 == current_begin_tag
        ):
            continue

        if i != begin:
            parsed_tags.append(
                Tag(
                    begin,
                    i - (i < num_tag_ids - 1),
                    idx2class((current_begin_tag + 1) // 2),
                )
            )

        if tag_id % 2 == 1:
            begin = i
            current_begin_tag = tag_id
        else:
            begin = i + 1
            current_begin_tag = -1
    return parsed_tags


def convert_global_pointer_to_tags(
    pointer: np.array,
    thresholds: Union[float, list[float]],
    idx2class: Callable[[int], str],
    add_start_end_tag: bool = False,
) -> list[Tag]:
    thresholds = _validate_thresholds(thresholds, pointer.shape[0])
    tags: list[Tag] = []
    for i, score_matrix in enumerate(pointer):
        label = idx2class(i)
        for start, end in zip(*np.where(expit(score_matrix) >= thresholds[i])):
            start -= add_start_end_tag
            end -= add_start_end_tag
            tags.append(Tag(start, end, label))
    return tags


def _compute_counts(
    label: Sequence[Tag], prediction: Sequence[Tag], classes: Sequence[str]
) -> list[tuple[str, int, int, int]]:
    truth_tags = set(label)
    prediction_tags = set(prediction)
    correct_tags = truth_tags & prediction_tags
    counts = [
        (
            cls,
            len([i for i, tag in enumerate(correct_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(truth_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(prediction_tags) if tag.label == cls]),
        )
        for cls in classes
    ]
    counts.append(("avg", len(correct_tags), len(truth_tags), len(prediction_tags)))
    return counts


def tagging_fscore(
    y: Sequence[Sequence[Tag]],
    p: Sequence[Sequence[Tag]],
    classes: Sequence[str],
) -> pd.DataFrame:
    df = pd.DataFrame(
        itertools.chain.from_iterable(
            _compute_counts(yi, pi, classes) for yi, pi in zip(y, p)
        ),
        columns=["class", "correct", "support", "prediction"],
    )
    df = df.groupby("class").sum()
    df["precision"] = df.correct / df.prediction
    df["recall"] = df.correct / df.support
    df["fscore"] = 2 * df.precision * df.recall / (df.precision + df.recall)
    df.fillna(0, inplace=True)
    df.reset_index(inplace=True)
    return df[["class", "precision", "recall", "fscore", "support"]]


from __future__ import annotations
from typing import Any
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="sknlp")
class MultiLabelCategoricalCrossentropy(tf.keras.losses.Loss):
    """
    苏剑林. (Apr. 25, 2020). 《将“softmax+交叉熵”推广到多标签分类问题 》[Blog post]. Retrieved from https://spaces.ac.cn/archives/7359
    """

    def __init__(
        self,
        reduction: str = tf.keras.losses.Reduction.AUTO,
        flatten_axis: int = 1,
        name: str = "multilabel_categorical_crossentropy",
    ) -> None:
        super().__init__(reduction=reduction, name=name)
        self.flatten_axis = flatten_axis

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        if isinstance(y_pred, tf.RaggedTensor):
            y_pred = y_pred.to_tensor(y_pred.dtype.min)
        else:
            y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)

        sample_size = tf.reduce_prod(tf.shape(y_true)[: self.flatten_axis])
        y_true = tf.reshape(y_true, (sample_size, -1))
        y_pred = tf.reshape(y_pred, (sample_size, -1))
        y_pred = (1 - 2 * y_true) * y_pred
        y_pred_neg: tf.Tensor = y_pred - y_true * y_pred.dtype.max
        y_pred_pos: tf.Tensor = y_pred - (1 - y_true) * y_pred.dtype.max
        zeros = tf.zeros_like(y_pred[..., :1])  # 用于生成logsum中的1
        y_pred_neg = tf.concat([y_pred_neg, zeros], axis=-1)
        y_pred_pos = tf.concat([y_pred_pos, zeros], axis=-1)
        neg_loss = tf.math.reduce_logsumexp(y_pred_neg, axis=-1)
        pos_loss = tf.math.reduce_logsumexp(y_pred_pos, axis=-1)
        return neg_loss + pos_loss

    def get_config(self) -> dict[str, Any]:
        return {**super().get_config(), "flatten_axis": self.flatten_axis}



from __future__ import annotations
from typing import Any
import sys

import tensorflow as tf
from tensorflow.python.ops.array_ops import boolean_mask

from .embedding_layer import SinusoidalPositionEmbedding


@tf.keras.utils.register_keras_serializable(package="sknlp")
class GlobalPointerLayer(tf.keras.layers.Layer):
    """
    苏剑林. (May. 01, 2021). 《GlobalPointer：用统一的方式处理嵌套和非嵌套NER 》[Blog post]. Retrieved from https://spaces.ac.cn/archives/8373
    """

    def __init__(
        self,
        heads: int,
        head_size: int,
        max_sequence_length: int,
        name: str = "global_pointer",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.heads = heads
        self.head_size = head_size
        self.max_sequence_length = max_sequence_length

    def build(self, input_shape: tf.TensorShape):
        self.dense = tf.keras.layers.Dense(self.head_size * self.heads * 2)
        self.sinusoial_postion = SinusoidalPositionEmbedding(
            self.head_size, self.max_sequence_length
        )
        super().build(input_shape)

    def call(self, inputs: tf.Tensor, mask: tf.Tensor, *args, **kwargs) -> tf.Tensor:
        """
        实现参照:https://github.com/bojone/bert4keras/blob/b3c28c3a7219c5a152082b1159af0e828923364d/bert4keras/layers.py#L1158
        """
        # (batch, seq, dim) -> (batch, seq, heads * head_size * 2)
        inputs = self.dense(inputs)
        # heads * (batch, seq, head_size * 2)
        inputs = tf.split(inputs, self.heads, axis=-1)
        # (batch, seq, heads, head_size * 2)
        inputs = tf.stack(inputs, axis=-2)
        # (batch, seq, heads, head_size)
        qw, kw = inputs[..., : self.head_size], inputs[..., self.head_size :]

        input_shape = tf.shape(inputs)
        # (1, seq)
        position_ids = tf.range(0, input_shape[1], dtype=tf.int64)[None]
        # (batch, seq)
        position_ids = tf.repeat(position_ids, input_shape[0], axis=0)
        # (bacth, seq, head_size)
        position_embedding = self.sinusoial_postion(position_ids)
        # (batch, seq, 1, head_size)
        sin_position = tf.repeat(position_embedding[..., None, ::2], 2, -1)
        cos_position = tf.repeat(position_embedding[..., None, 1::2], 2, -1)
        # (batch, seq, heads, head_size / 2, 2)
        qw2 = tf.stack([-qw[..., 1::2], qw[..., ::2]], axis=4)
        qw2 = tf.reshape(qw2, tf.shape(qw))
        qw = qw * cos_position + qw2 * sin_position
        kw2 = tf.stack([-kw[..., 1::2], kw[..., ::2]], axis=4)
        kw2 = tf.reshape(kw2, tf.shape(kw))
        kw = kw * cos_position + kw2 * sin_position

        logits = tf.einsum("bmhs,bnhs->bhmn", qw, kw)
        mask = tf.cast(mask, logits.dtype)
        mask = tf.expand_dims(tf.expand_dims(mask, 1), 3) * tf.expand_dims(
            tf.expand_dims(mask, 1), 1
        )
        logits = logits * mask + logits.dtype.min * (1 - mask)
        boolean_mask = tf.repeat(tf.cast(mask, tf.bool), self.heads, axis=1)
        # 排除下三角
        mask = tf.linalg.band_part(tf.ones_like(logits), 0, -1)
        logits = logits * mask + logits.dtype.min * (1 - mask)
        return tf.ragged.boolean_mask(logits / self.head_size ** 0.5, boolean_mask)

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        return tf.TensorShape([input_shape[0], self.heads, input_shape[1] ** 2])

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "heads": self.heads,
            "head_size": self.head_size,
            "max_sequence_length": self.max_sequence_length,
        }


from __future__ import annotations
from typing import Sequence, Optional, Tuple, Callable, Any

import pandas as pd

import tensorflow as tf

from .classification_dataset import ClassificationDataset
from .tagging_dataset import TaggingDataset
from .similarity_dataset import SimilarityDataset


class BertClassificationDataset(ClassificationDataset):
    def __init__(
        self,
        tokenizer: Callable[[str], list[int]],
        labels: Sequence[str],
        X: Optional[Sequence[Any]] = None,
        y: Optional[Sequence[Any]] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        is_multilabel: bool = True,
        max_length: Optional[int] = None,
    ):
        super().__init__(
            tokenizer,
            labels,
            X=X,
            y=y,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            is_multilabel=is_multilabel,
            max_length=max_length,
            text_dtype=tf.string,
            label_dtype=tf.float32,
        )

    @property
    def batch_padding_shapes(self) -> Optional[list[Tuple]]:
        return None

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("UTF-8")[: self.max_length]


class BertTaggingDataset(TaggingDataset):
    def __init__(
        self,
        tokenizer: Callable[[str], list[int]],
        labels: Sequence[str],
        X: Optional[Sequence[Any]] = None,
        y: Optional[Sequence[Any]] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        add_start_end_tag: bool = True,
        output_format: str = "global_pointer",
        max_length: Optional[int] = None,
    ):
        super().__init__(
            tokenizer,
            labels,
            X=X,
            y=y,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            output_format=output_format,
            add_start_end_tag=add_start_end_tag,
            max_length=max_length,
            text_dtype=tf.string,
            label_dtype=tf.int32,
        )

    @property
    def batch_padding_shapes(self) -> Optional[list[Tuple]]:
        if self.output_format == "bio":
            return ((), (None,))
        else:
            return ((), (None, None, None))

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("UTF-8")[: self.max_length]


class BertSimilarityDataset(SimilarityDataset):
    def __init__(
        self,
        tokenizer: Callable[[str], list[int]],
        labels: Sequence[str],
        X: Optional[Sequence[Any]] = None,
        y: Optional[Sequence[Any]] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        max_length: Optional[int] = None,
    ):
        super().__init__(
            tokenizer,
            labels,
            X=X,
            y=y,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            max_length=max_length,
            text_dtype=tf.string,
            label_dtype=tf.float32,
        )

    @property
    def batch_padding_shapes(self) -> Optional[list[Tuple]]:
        return None

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("UTF-8")[: self.max_length]


from __future__ import annotations

try:
    import jieba_fast as jieba
except ImportError:
    import jieba
import tensorflow as tf

from sknlp.vocab import Vocab
from sknlp.layers import BertPreprocessingLayer


class Tokenizer:
    def __init__(self, vocab: Vocab) -> None:
        self.vocab = vocab

    def tokenize(self, text: str) -> list[int]:
        raise NotImplementedError()


class JiebaTokenizer(Tokenizer):
    def tokenize(self, text: str) -> list[int]:
        tokens = jieba.lcut(text, HMM=False)
        return self.vocab.token2idx(tokens)


class CharTokenizer(Tokenizer):
    def tokenize(self, text: str) -> list[int]:
        tokens = list(text)
        return self.vocab.token2idx(tokens)


class BertTokenizer(Tokenizer):
    def __init__(self, vocab: Vocab) -> None:
        super().__init__(vocab)
        layer = BertPreprocessingLayer(self.vocab.sorted_tokens)
        layer.build(tf.TensorShape([None]))
        self._tokenizer = layer.tokenizer

    def tokenize(self, text: str) -> list[int]:
        token_ids: tf.RaggedTensor = self._tokenizer.tokenize([text])
        token_ids = token_ids.merge_dims(-2, -1)
        return token_ids.to_list()[0]


def get_tokenizer(name: str, vocab: Vocab) -> Tokenizer:
    if name == "jieba":
        return JiebaTokenizer(vocab)
    elif name == "list" or name == "char":
        return CharTokenizer(vocab)
    elif name is None:
        return BertTokenizer(vocab)
    else:
        raise ValueError("unknown segmenter %s" % name)


from __future__ import annotations

from tensorflow.keras.layers import Bidirectional, Dropout

from sknlp.layers import LSTMP
from .deep_classifier import DeepClassifier


class TextRNNClassifier(DeepClassifier):
    def __init__(
        self,
        classes: list[str],
        is_multilabel: bool = True,
        max_sequence_length: int = 100,
        dropout: float = 0.5,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 128,
        fc_activation: str = "tanh",
        num_rnn_layers: int = 1,
        rnn_hidden_size: int = 512,
        rnn_projection_size: int = 128,
        rnn_recurrent_clip: float = 3.0,
        rnn_projection_clip: float = 3.0,
        rnn_kernel_initializer="glorot_uniform",
        rnn_recurrent_initializer="orthogonal",
        rnn_projection_initializer="glorot_uniform",
        rnn_bias_initializer="zeros",
        rnn_kernel_regularizer=None,
        rnn_recurrent_regularizer=None,
        rnn_projection_regularizer=None,
        rnn_bias_regularizer=None,
        rnn_kernel_constraint=None,
        rnn_recurrent_constraint=None,
        rnn_projection_constraint=None,
        rnn_bias_constraint=None,
        text2vec=None,
        **kwargs
    ):
        super().__init__(
            classes,
            is_multilabel=is_multilabel,
            max_sequence_length=max_sequence_length,
            num_fc_layers=num_fc_layers,
            fc_hidden_size=fc_hidden_size,
            fc_activation=fc_activation,
            text2vec=text2vec,
            algorithm="rnn",
            **kwargs
        )
        self.dropout = dropout
        self.num_rnn_layers = num_rnn_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.rnn_projection_size = rnn_projection_size
        self.rnn_recurrent_clip = rnn_recurrent_clip
        self.rnn_projection_clip = rnn_projection_clip
        self.rnn_kernel_initializer = rnn_kernel_initializer
        self.rnn_recurrent_initializer = rnn_recurrent_initializer
        self.rnn_projection_initializer = rnn_projection_initializer
        self.rnn_bias_initializer = rnn_bias_initializer
        self.rnn_kernel_regularizer = rnn_kernel_regularizer
        self.rnn_recurrent_regularizer = rnn_recurrent_regularizer
        self.rnn_projection_regularizer = rnn_projection_regularizer
        self.rnn_bias_regularizer = rnn_bias_regularizer
        self.rnn_kernel_constraint = rnn_kernel_constraint
        self.rnn_recurrent_constraint = rnn_recurrent_constraint
        self.rnn_projection_constraint = rnn_projection_constraint
        self.rnn_bias_constraint = rnn_bias_constraint

    def build_encode_layer(self, inputs):
        outputs = self.text2vec(inputs)
        for i in range(self.num_rnn_layers):
            return_sequences = i != self.num_rnn_layers - 1
            outputs = Bidirectional(
                LSTMP(
                    self.rnn_hidden_size,
                    projection_size=self.rnn_projection_size,
                    recurrent_clip=self.rnn_recurrent_clip,
                    projection_clip=self.rnn_projection_clip,
                    dropout=self.dropout,
                    recurrent_dropout=self.dropout,
                    kernel_initializer=self.rnn_kernel_initializer,
                    recurrent_initializer=self.rnn_recurrent_initializer,
                    projection_initializer=self.rnn_projection_initializer,
                    bias_initializer=self.rnn_bias_initializer,
                    kernel_regularizer=self.rnn_kernel_regularizer,
                    recurrent_regularizer=self.rnn_recurrent_regularizer,
                    projection_regularizer=self.rnn_projection_regularizer,
                    bias_regularizer=self.rnn_bias_regularizer,
                    kernel_constraint=self.rnn_kernel_constraint,
                    recurrent_constraint=self.rnn_recurrent_constraint,
                    projection_constraint=self.rnn_projection_constraint,
                    bias_constraint=self.rnn_bias_constraint,
                    return_sequences=return_sequences,
                )
            )(outputs)
        if self.dropout:
            outputs = Dropout(self.dropout, noise_shape=None, name="encoder_dropout")(
                outputs
            )
        return outputs


from __future__ import annotations
from typing import Any
import tensorflow as tf
import numpy as np


@tf.keras.utils.register_keras_serializable(package="sknlp")
class SinusoidalPositionEmbedding(tf.keras.layers.Layer):
    def __init__(
        self,
        output_size: int,
        max_sequence_length: int,
        name: str = "sinusoidal_position_embedding",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.output_size = output_size
        self.max_sequence_length = max_sequence_length

    def build(self, input_shape: tf.TensorShape) -> None:
        self.embedding = tf.keras.layers.Embedding(
            self.max_sequence_length,
            self.output_size,
        )
        self.embedding.build(input_shape)
        weights = np.array(
            [
                [
                    pos / np.power(10000, 2 * (j // 2) / self.output_size)
                    for j in range(self.output_size)
                ]
                for pos in range(self.max_sequence_length)
            ]
        )
        weights[:, ::2] = np.sin(weights[:, ::2])
        weights[:, 1::2] = np.cos(weights[:, 1::2])
        self.embedding.set_weights([weights])
        self.trainable = False
        super().build(input_shape)

    def call(self, inputs: tf.Tensor, *args, **kwargs) -> None:
        return self.embedding(inputs)

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        return tf.TensorShape([*input_shape.as_list(), self.output_size])

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "output_size": self.output_size,
            "max_sequence_length": self.max_sequence_length,
        }


from __future__ import annotations
from typing import Optional, Callable

import tensorflow as tf

from .weight_decay_scheduler import WeightDecayScheduler
from .learning_rate_scheduler import LearningRateScheduler

from .tagging_fscore_metric import TaggingFScoreMetric


def create_exponential_decay_scheduler(
    update_factor: float,
    update_epochs: int,
) -> Callable[[int, float], float]:
    def scheduler(epoch: int, x: float) -> float:
        return x * update_factor ** (epoch % update_epochs == 0 and epoch != 0)

    return scheduler


def create_warmup_scheduler(warmup_steps: int) -> Callable[[int, float], float]:
    def scheduler(step: int, x: float) -> float:
        if step == 0 and warmup_steps > 0:
            return x / warmup_steps
        if step >= warmup_steps:
            return x
        return x * (step + 1) / step

    return scheduler


def default_supervised_model_callbacks(
    learning_rate_update_factor: float,
    learning_rate_update_epochs: int,
    learning_rate_warmup_steps: int,
    use_weight_decay: bool = False,
    has_validation_dataset: bool = False,
    enable_early_stopping: bool = False,
    early_stopping_monitor: str = "val_loss",
    early_stopping_monitor_direction: str = "min",
    early_stopping_patience: Optional[int] = None,
    early_stopping_min_delta: float = 0.0,
    early_stopping_use_best_epoch: bool = False,
    checkpoint: Optional[str] = None,
    log_file: str = None,
    verbose: int = 2,
) -> list[tf.keras.callbacks.Callback]:
    decay_scheduler = create_exponential_decay_scheduler(
        learning_rate_update_factor,
        learning_rate_update_epochs,
    )
    warmup_scheduler = create_warmup_scheduler(learning_rate_warmup_steps)
    learning_rate_decay = LearningRateScheduler(
        warmup_scheduler, decay_scheduler, verbose=verbose
    )
    callbacks = [learning_rate_decay]
    if use_weight_decay > 0:
        callbacks.append(
            WeightDecayScheduler(warmup_scheduler, decay_scheduler, verbose=verbose)
        )

    if enable_early_stopping and has_validation_dataset:
        callbacks.append(
            tf.keras.callbacks.EarlyStopping(
                monitor=early_stopping_monitor,
                min_delta=early_stopping_min_delta,
                patience=early_stopping_patience or learning_rate_update_epochs,
                mode=early_stopping_monitor_direction,
                restore_best_weights=early_stopping_use_best_epoch,
            )
        )
    if checkpoint is not None:
        callbacks.append(
            tf.keras.callbacks.ModelCheckpoint(
                checkpoint,
                monitor=early_stopping_monitor,
                mode=early_stopping_monitor_direction,
                save_best_only=has_validation_dataset,
                verbose=verbose,
            )
        )
    if log_file is not None:
        callbacks.append(tf.keras.callbacks.CSVLogger(log_file))
    return callbacks


__all__ = [
    "WeightDecayScheduler",
    "LearningRateSchduler",
    "TaggingFScoreMetric",
    "default_supervised_model_callbacks",
]


import sys

import tensorflow as tf


def logits2pred(logits: tf.Tensor, activation: str) -> tf.Tensor:
    if isinstance(logits, tf.RaggedTensor):
        fill_value = 0
        if activation in ("sigmoid", "softmax"):
            fill_value = logits.dtype.min
        logits = logits.to_tensor(fill_value)
    return tf.keras.activations.get(activation)(logits)


from typing import Any, Dict, Sequence, Optional

import tensorflow as tf
from tensorflow.keras.layers import Dropout

from sknlp.layers import BertPairPreprocessingLayer
from sknlp.data import BertSimilarityDataset
from sknlp.module.text2vec import Bert2vec
from .deep_discriminator import DeepDiscriminator


class BertDiscriminator(DeepDiscriminator):
    dataset_class = BertSimilarityDataset

    def __init__(
        self,
        classes: Sequence[str] = ("相似度",),
        max_sequence_length: int = 120,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        fc_activation: str = "tanh",
        dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ) -> None:
        super().__init__(
            classes=classes,
            algorithm="bert",
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            **kwargs
        )
        self.num_fc_layers = num_fc_layers
        self.fc_hidden_size = fc_hidden_size
        self.fc_activation = fc_activation
        self.dropout = dropout
        self.inputs = [
            tf.keras.Input(shape=(), dtype=tf.string, name="text_input"),
            tf.keras.Input(shape=(), dtype=tf.string, name="context_input"),
        ]

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        preprocessing_layer = BertPairPreprocessingLayer(
            self.text2vec.vocab.sorted_tokens
        )
        token_ids, type_ids = preprocessing_layer(inputs)
        if self.dropout:
            self.text2vec.update_dropout(dropout=self.dropout)
        _, cls = self.text2vec([token_ids, type_ids])
        if self.dropout:
            return Dropout(self.dropout, name="embedding_dropout")(cls)
        return cls

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "dropout": self.dropout}


from __future__ import annotations
from typing import Any, Sequence, Optional

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Dropout

from sknlp.layers import BertCharPreprocessingLayer
from sknlp.data import BertTaggingDataset
from sknlp.module.text2vec import Bert2vec
from .deep_tagger import DeepTagger


class BertTagger(DeepTagger):
    dataset_class = BertTaggingDataset

    def __init__(
        self,
        classes: Sequence[str],
        output_format: str = "global_pointer",
        crf_learning_rate_multiplier: float = 1.0,
        max_sequence_length: int = 120,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        fc_activation: str = "tanh",
        dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ) -> None:
        super().__init__(
            classes,
            add_start_end_tag=True,
            output_format=output_format,
            crf_learning_rate_multiplier=crf_learning_rate_multiplier,
            algorithm="bert",
            num_fc_layers=num_fc_layers,
            fc_hidden_size=fc_hidden_size,
            fc_activation=fc_activation,
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            **kwargs
        )
        self.dropout = dropout
        if self.output_format == "bio":
            self.inputs = [
                tf.keras.Input(shape=(), dtype=tf.string, name="text_input"),
                tf.keras.Input(shape=(None,), dtype=tf.int32, name="tag_id"),
            ]
        else:
            self.inputs = tf.keras.Input(shape=(), dtype=tf.string, name="text_input")

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        if self.output_format == "bio":
            texts = inputs[0]
        else:
            texts = inputs
        token_ids = BertCharPreprocessingLayer(self.text2vec.vocab.sorted_tokens)(texts)

        mask = tf.keras.layers.Lambda(
            lambda x: tf.cast(x != 0, tf.int32), name="mask_layer"
        )(token_ids)
        if self.dropout:
            self.text2vec.update_dropout(dropout=self.dropout)
        embeddings, _ = self.text2vec(
            [token_ids, K.zeros_like(token_ids, dtype=tf.int64)]
        )
        if self.dropout:
            noise_shape = (None, 1, self.text2vec.embedding_size)
            embeddings = Dropout(
                self.dropout,
                noise_shape=noise_shape,
                name="encoder_dropout",
            )(embeddings)

        if self.output_format == "bio":
            return embeddings, mask, inputs[1]
        else:
            return embeddings, mask

    def get_config(self) -> dict[str, Any]:
        return {**super().get_config(), "dropout": self.dropout}

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "BertTagger":
        config.pop("add_start_end_tag")
        return super().from_config(config)


from typing import Optional

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras.layers import InputSpec
from tensorflow.python.keras.layers.recurrent import LSTM, LSTMCell
from tensorflow.python.training.tracking import data_structures

from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint


@tf.keras.utils.register_keras_serializable(package="sknlp")
class LSTMPCell(LSTMCell):
    """
    Long-Short Term Memory Projected (LSTMP) network cell with cell clip.
    (https://arxiv.org/abs/1402.1128)

    Parameters
    ----------
    units: Positive integer, dimensionality of the hidden state.
    projection_size: int, dimensionality of the output.
    activation: Activation function to use. Default: hyperbolic tangent
      (`tanh`). If you pass `None`, no activation is applied (ie. "linear"
      activation: `a(x) = x`).
    recurrent_activation: Activation function to use for the recurrent step.
      Default: sigmoid (`sigmoid`). If you pass `None`,
      no activation is applied (ie. "linear" activation: `a(x) = x`).
    use_bias: Boolean, whether the layer uses a bias vector.
    kernel_initializer: Initializer for the `kernel` weights matrix, used for
      the linear transformation of the inputs.
    recurrent_initializer: Initializer for the `recurrent_kernel` weights
      matrix, used for the linear transformation of the recurrent state.
    projection_initializer: Initializer for the `projection_kernel` weights
      matrix, used for the linear transformation of the hidden state.
    bias_initializer: Initializer for the bias vector.
    unit_forget_bias: Boolean. If True, add 1 to the bias of the forget gate at
      initialization. Setting it to true will also force
      `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
        al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
    kernel_regularizer: Regularizer function applied to the `kernel` weights
      matrix.
    recurrent_regularizer: Regularizer function applied to
      the `recurrent_kernel` weights matrix.
    projection_regularizer: Regularizer function applied to
      the `projection_kernel` weights matrix.
    bias_regularizer: Regularizer function applied to the bias vector.
    kernel_constraint: Constraint function applied to the `kernel` weights
      matrix.
    recurrent_constraint: Constraint function applied to the `recurrent_kernel`
      weights matrix.
    projection_constraint: Constraint function applied to the
      `projection_kernel` weights matrix.
    bias_constraint: Constraint function applied to the bias vector.
    dropout: Float between 0 and 1. Fraction of the units to drop for the
      linear transformation of the inputs.
    recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
      the linear transformation of the recurrent state.
    projection_dropout: Float between 0 and 1. Fraction of the units to drop
      for the linear transformation of the hidden state.
    implementation: Implementation mode, either 1 or 2.
      Mode 1 will structure its operations as a larger number of smaller dot
      products and additions, whereas mode 2 (default) will batch them into
      fewer, larger operations. These modes will have different performance
      profiles on different hardware and for different applications.

    Inputs
    ----------
    inputs: 2D tensor with shape: `(batch_size, input_length)`.
    states: List of state tensors corresponding to the previous timestep.
    training: Python boolean indicating whether the layer should behave in
      training mode or in inference mode. Only relevant when `dropout`,
      `recurrent_dropout` or `projection_dropout`is used.
    """

    def __init__(
        self,
        units: int,
        projection_size: int = 100,
        activation: str = "tanh",
        recurrent_activation: str = "hard_sigmoid",
        use_bias: bool = True,
        kernel_initializer: WeightInitializer = "glorot_uniform",
        recurrent_initializer: WeightInitializer = "orthogonal",
        projection_initializer: WeightInitializer = "glorot_uniform",
        bias_initializer: WeightInitializer = "zeros",
        unit_forget_bias: bool = True,
        kernel_regularizer: Optional[WeightRegularizer] = None,
        recurrent_regularizer: Optional[WeightRegularizer] = None,
        projection_regularizer: Optional[WeightRegularizer] = None,
        bias_regularizer: Optional[WeightRegularizer] = None,
        kernel_constraint: Optional[WeightConstraint] = None,
        recurrent_constraint: Optional[WeightConstraint] = None,
        projection_constraint: Optional[WeightConstraint] = None,
        bias_constraint: Optional[WeightConstraint] = None,
        dropout: float = 0.0,
        recurrent_dropout: float = 0.0,
        recurrent_clip: Optional[float] = None,
        projection_clip: Optional[float] = None,
        implementation: int = 2,
        **kwargs
    ) -> None:
        super().__init__(
            units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            unit_forget_bias=unit_forget_bias,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            implementation=implementation,
            **kwargs
        )
        self.projection_size = projection_size
        self.recurrent_clip = recurrent_clip
        self.projection_clip = projection_clip
        self.projection_initializer = initializers.get(projection_initializer)
        self.projection_regularizer = regularizers.get(projection_regularizer)
        self.projection_constraint = constraints.get(projection_constraint)
        self.state_size = data_structures.NoDependency(
            [self.projection_size, self.units]
        )
        self.output_size = self.projection_size

    def build(self, input_shape):
        input_dim = input_shape[-1]
        self.kernel = self.add_weight(
            shape=(input_dim, self.units * 4),
            name="kernel",
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        self.recurrent_kernel = self.add_weight(
            shape=(self.projection_size, self.units * 4),
            name="recurrent_kernel",
            initializer=self.recurrent_initializer,
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint,
        )

        if self.use_bias:
            if self.unit_forget_bias:

                def bias_initializer(_, *args, **kwargs):
                    return K.concatenate(
                        [
                            self.bias_initializer((self.units,), *args, **kwargs),
                            initializers.Ones()((self.units,), *args, **kwargs),
                            self.bias_initializer((self.units * 2,), *args, **kwargs),
                        ]
                    )

            else:
                bias_initializer = self.bias_initializer
            self.bias = self.add_weight(
                shape=(self.units * 4,),
                name="bias",
                initializer=bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = None

        self.projection_kernel = self.add_weight(
            shape=(self.units, self.projection_size),
            name="projection_kernel",
            initializer=self.projection_initializer,
            regularizer=self.projection_regularizer,
            constraint=self.projection_constraint,
        )
        self.built = True

    def call(self, inputs, states, training=None):
        h, (_, c) = super().call(inputs, states, training=training)
        if self.recurrent_clip is not None:
            c = K.clip(c, -self.recurrent_clip, self.recurrent_clip)
        r = K.dot(h, self.projection_kernel)
        if self.projection_clip is not None:
            r = K.clip(r, -self.projection_clip, self.projection_clip)
        return r, [r, c]

    def get_config(self):
        config = {
            "projection_size": self.projection_size,
            "recurrent_clip": self.recurrent_clip,
            "projection_clip": self.projection_clip,
            "projection_initializer": initializers.serialize(
                self.projection_initializer
            ),
            "projection_regularizer": regularizers.serialize(
                self.projection_regularizer
            ),
            "projection_constraint": constraints.serialize(self.projection_constraint),
        }
        base_config = super().get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package="sknlp")
class LSTMP(LSTM):
    """

    Parameters
    ----------
    units: Positive integer, dimensionality of the hidden state.
    projection_size: int, dimensionality of the output.
    activation: Activation function to use. Default: hyperbolic tangent
      (`tanh`). If you pass `None`, no activation is applied (ie. "linear"
      activation: `a(x) = x`).
    recurrent_activation: Activation function to use for the recurrent step.
      Default: sigmoid (`sigmoid`). If you pass `None`,
      no activation is applied (ie. "linear" activation: `a(x) = x`).
    use_bias: Boolean, whether the layer uses a bias vector.
    kernel_initializer: Initializer for the `kernel` weights matrix, used for
      the linear transformation of the inputs.
    recurrent_initializer: Initializer for the `recurrent_kernel` weights
      matrix, used for the linear transformation of the recurrent state.
    projection_initializer: Initializer for the `projection_kernel` weights
      matrix, used for the linear transformation of the hidden state.
    bias_initializer: Initializer for the bias vector.
    unit_forget_bias: Boolean. If True, add 1 to the bias of the forget gate at
      initialization. Setting it to true will also force
      `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
        al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
    kernel_regularizer: Regularizer function applied to the `kernel` weights
      matrix.
    recurrent_regularizer: Regularizer function applied to
      the `recurrent_kernel` weights matrix.
    projection_regularizer: Regularizer function applied to
      the `projection_kernel` weights matrix.
    bias_regularizer: Regularizer function applied to the bias vector.
    kernel_constraint: Constraint function applied to the `kernel` weights
      matrix.
    recurrent_constraint: Constraint function applied to the `recurrent_kernel`
      weights matrix.
    projection_constraint: Constraint function applied to the
      `projection_kernel` weights matrix.
    bias_constraint: Constraint function applied to the bias vector.
    dropout: Float between 0 and 1. Fraction of the units to drop for the
      linear transformation of the inputs.
    recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
      the linear transformation of the recurrent state.
    projection_dropout: Float between 0 and 1. Fraction of the units to drop
      for the linear transformation of the hidden state.
    implementation: Implementation mode, either 1 or 2.
      Mode 1 will structure its operations as a larger number of smaller dot
      products and additions, whereas mode 2 (default) will batch them into
      fewer, larger operations. These modes will have different performance
      profiles on different hardware and for different applications.
    return_sequences: Boolean. Whether to return the last output.
      in the output sequence, or the full sequence.
    return_state: Boolean. Whether to return the last state
      in addition to the output.
    go_backwards: Boolean (default False).
      If True, process the input sequence backwards and return the
      reversed sequence.
    stateful: Boolean (default False). If True, the last state
      for each sample at index i in a batch will be used as initial
      state for the sample of index i in the following batch.
    unroll: Boolean (default False).
      If True, the network will be unrolled,
      else a symbolic loop will be used.
      Unrolling can speed-up a RNN,
      although it tends to be more memory-intensive.
      Unrolling is only suitable for short sequences.
    time_major: The shape format of the `inputs` and `outputs` tensors.
      If True, the inputs and outputs will be in shape
      `(timesteps, batch, ...)`, whereas in the False case, it will be
      `(batch, timesteps, ...)`. Using `time_major = True` is a bit more
      efficient because it avoids transposes at the beginning and end of the
      RNN calculation. However, most TensorFlow data is batch-major, so by
      default this function accepts input and emits output in batch-major
      form.

    Inputs
    ----------
    inputs: 2D tensor with shape: `(batch_size, input_length)`.
    mask: Binary tensor of shape `(samples, timesteps)` indicating whether
      a given timestep should be masked.
    training: Python boolean indicating whether the layer should behave in
      training mode or in inference mode. Only relevant when `dropout`,
      `recurrent_dropout` or `projection_dropout`is used.
    initial_states: List of state tensors corresponding to the
      previous timestep.
    """

    def __init__(
        self,
        units,
        projection_size: int = 100,
        activation: str = "tanh",
        recurrent_activation: str = "sigmoid",
        use_bias: bool = True,
        kernel_initializer: WeightInitializer = "glorot_uniform",
        recurrent_initializer: WeightInitializer = "orthogonal",
        projection_initializer: WeightInitializer = "glorot_uniform",
        bias_initializer: WeightInitializer = "zeros",
        unit_forget_bias: bool = True,
        kernel_regularizer: Optional[WeightRegularizer] = None,
        recurrent_regularizer: Optional[WeightRegularizer] = None,
        projection_regularizer: Optional[WeightRegularizer] = None,
        bias_regularizer: Optional[WeightRegularizer] = None,
        kernel_constraint: Optional[WeightConstraint] = None,
        recurrent_constraint: Optional[WeightConstraint] = None,
        projection_constraint: Optional[WeightConstraint] = None,
        bias_constraint: Optional[WeightConstraint] = None,
        dropout: float = 0.0,
        recurrent_dropout: float = 0.0,
        recurrent_clip: Optional[float] = None,
        projection_clip: Optional[float] = None,
        implementation: int = 2,
        activity_regularizer: Optional[WeightRegularizer] = None,
        return_sequences: bool = False,
        return_state: bool = False,
        go_backwards: bool = False,
        stateful: bool = False,
        unroll: bool = False,
        **kwargs
    ):
        cell = LSTMPCell(
            units,
            projection_size=projection_size,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            projection_initializer=projection_initializer,
            unit_forget_bias=unit_forget_bias,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            projection_regularizer=projection_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            projection_constraint=projection_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            recurrent_clip=recurrent_clip,
            projection_clip=projection_clip,
            implementation=implementation,
            dtype=kwargs.get("dtype"),
        )
        super(LSTM, self).__init__(
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            unroll=unroll,
            **kwargs
        )
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.supports_masking = True
        self.input_spec = [InputSpec(ndim=3)]

    @property
    def projection_size(self):
        return self.cell.projection_size

    @property
    def projection_initializer(self):
        return initializers.serialize(self.cell.projection_initializer)

    @property
    def projection_regularizer(self):
        return regularizers.serialize(self.cell.projection_regularizer)

    @property
    def projection_constraint(self):
        return constraints.serialize(self.cell.projection_constraint)

    @property
    def recurrent_clip(self):
        return self.cell.recurrent_clip

    @property
    def projection_clip(self):
        return self.cell.projection_clip

    def compute_mask(self, inputs, mask):
        output_mask = mask
        if self.return_state:
            state_mask = [None for _ in self.states]
            return [output_mask] + state_mask
        else:
            return output_mask

    def get_config(self):
        base_config = super().get_config()
        return {**base_config, **self.cell.get_config()}

    @classmethod
    def from_config(cls, config):
        if "implementation" in config and config["implementation"] == 0:
            config["implementation"] = 2
        return cls(**config)


from __future__ import annotations
from typing import Any

import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense, InputSpec


@tf.keras.utils.register_keras_serializable(package="sknlp")
class MLPLayer(Layer):
    def __init__(
        self,
        num_layers: int,
        hidden_size: int = 256,
        output_size: int = 1,
        activation: str = "tanh",
        name: str = "mlp",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.activation = activation
        self.supports_masking = True

    def build(self, input_shape: tf.TensorShape) -> None:
        self.dense_layers = []
        for i in range(self.num_layers):
            if i == self.num_layers - 1:
                self.dense_layers.append(Dense(self.output_size, name="dense-%d" % i))
            else:
                self.dense_layers.append(
                    Dense(
                        self.hidden_size,
                        activation=tf.keras.activations.get(self.activation),
                        name="dense-%d" % i,
                    )
                )
        last_dim = input_shape[-1]
        self.input_spec = InputSpec(min_ndim=2, axes={-1: last_dim})
        super().build(input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        outputs = inputs
        for dense in self.dense_layers:
            outputs = dense(outputs)
        return outputs

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        input_shape = input_shape.with_rank_at_least(2)
        if input_shape[-1] is None:
            raise ValueError(
                "The innermost dimension of input_shape must be defined, "
                "but saw: %s" % input_shape
            )
        return input_shape[:-1].concatenate(self.output_size)

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "num_layers": self.num_layers,
            "hidden_size": self.hidden_size,
            "output_size": self.output_size,
            "activation": self.activation,
        }


from __future__ import annotations
from typing import Any, Optional

import tensorflow as tf
from tensorflow_addons.text import crf_log_likelihood, crf_decode
import tensorflow.keras.backend as K

from sknlp.typing import WeightInitializer


@tf.keras.utils.register_keras_serializable(package="sknlp")
class CrfLossLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        num_tags: int,
        learning_rate_multiplier: float = 1.0,
        max_sequence_length: int = 120,
        initializer: WeightInitializer = "Orthogonal",
        name: str = "crf",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.num_tags = num_tags
        self.learning_rate_multiplier = learning_rate_multiplier
        self.max_sequence_length = max_sequence_length
        self.initializer = initializer

    def build(self, input_shape: tf.TensorShape) -> None:
        super().build(input_shape)
        self.transition_weight = self.add_weight(
            shape=(self.num_tags, self.num_tags),
            dtype=tf.float32,
            initializer=tf.keras.initializers.get(self.initializer),
            name="transision_weight",
        )
        K.set_value(
            self.transition_weight,
            K.get_value(self.transition_weight) / self.learning_rate_multiplier,
        )

    def call(self, inputs: list[tf.Tensor], mask: tf.Tensor) -> tf.RaggedTensor:
        emissions, tag_ids = inputs
        mask = tf.cast(mask, tf.int32)
        sequence_lengths = tf.math.reduce_sum(mask, axis=1)
        likelihoods, _ = crf_log_likelihood(
            emissions,
            tag_ids,
            sequence_lengths,
            self.transition_weight * self.learning_rate_multiplier,
        )
        loss = tf.math.negative(tf.math.reduce_mean(likelihoods))
        self.add_loss(loss)

        decoded_tag_ids, _ = crf_decode(
            emissions,
            self.transition_weight * self.learning_rate_multiplier,
            sequence_lengths,
        )
        is_equal = tf.cast(tf.equal(tag_ids, decoded_tag_ids), tf.int32)
        tag_accuracy = tf.reduce_sum(is_equal * mask) / tf.reduce_sum(mask)
        self.add_metric(tag_accuracy, name="tag_accuracy")
        boolean_mask = tf.cast(mask, tf.bool)
        return tf.ragged.boolean_mask(decoded_tag_ids, boolean_mask)

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "num_tags": self.num_tags,
            "learning_rate_multiplier": self.learning_rate_multiplier,
            "max_sequence_length": self.max_sequence_length,
            "initializer": self.initializer,
        }


@tf.keras.utils.register_keras_serializable(package="sknlp")
class CrfDecodeLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        num_tags: int,
        learning_rate_multiplier: float = 1.0,
        max_sequence_length: int = 120,
        initializer: WeightInitializer = "Orthogonal",
        name: str = "crf",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.num_tags = num_tags
        self.learning_rate_multiplier = learning_rate_multiplier
        self.max_sequence_length = max_sequence_length
        self.initializer = initializer

    def build(self, input_shape: tf.TensorShape) -> None:
        super().build(input_shape)
        self.transition_weight = self.add_weight(
            shape=(self.num_tags, self.num_tags),
            dtype=tf.float32,
            initializer=tf.keras.initializers.get(self.initializer),
            name="transision_weight",
        )

    def call(self, emissions: tf.Tensor, mask: tf.Tensor) -> tf.Tensor:
        mask = tf.cast(mask, tf.int32)
        sequence_lengths = tf.math.reduce_sum(mask, axis=1)
        decoded_tag_ids, _ = crf_decode(
            emissions,
            self.transition_weight * self.learning_rate_multiplier,
            sequence_lengths,
        )
        return decoded_tag_ids * mask

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "num_tags": self.num_tags,
            "learning_rate_multiplier": self.learning_rate_multiplier,
            "max_sequence_length": self.max_sequence_length,
            "initializer": self.initializer,
        }



from __future__ import annotations
from typing import Sequence, Union, Optional, Any

import os
import logging

import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np
import pandas as pd
from tabulate import tabulate

from sknlp.data import NLPDataset
from sknlp.vocab import Vocab
from sknlp.callbacks import default_supervised_model_callbacks
from .text2vec import Text2vec
from .base_model import BaseNLPModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class SupervisedNLPModel(BaseNLPModel):
    dataset_class = NLPDataset
    dataset_args = []

    def __init__(
        self,
        classes: list[str],
        max_sequence_length: Optional[int] = None,
        text2vec: Optional[Text2vec] = None,
        task: Optional[str] = None,
        algorithm: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(max_sequence_length=max_sequence_length, **kwargs)
        if text2vec is not None:
            self.text2vec = text2vec

        self._task = task
        self._algorithm = algorithm
        self._class2idx = dict(zip(classes, range(len(classes))))
        self._idx2class = dict(zip(range(len(classes)), classes))

    @property
    def dataset_kwargs(self) -> dict[str, Any]:
        kwargs = dict()
        for field in self.dataset_args:
            instance_field = getattr(self, field, None)
            if instance_field is not None:
                kwargs[field] = instance_field
        return kwargs

    @property
    def num_classes(self) -> int:
        return len(self._class2idx)

    @property
    def classes(self) -> list[str]:
        return list(self._class2idx.keys())

    @property
    def text2vec(self) -> Text2vec:
        return self._text2vec

    @text2vec.setter
    def text2vec(self, tv: Text2vec) -> None:
        if self.max_sequence_length is not None and tv.max_sequence_length is not None:
            self._max_sequence_length = min(
                self.max_sequence_length, tv.max_sequence_length
            )
        else:
            self._max_sequence_length = (
                self.max_sequence_length or tv.max_sequence_length
            )
        self._sequence_length = tv.sequence_length
        self._segmenter = tv.segmenter
        self._text2vec = tv

    @property
    def training_dataset(self) -> NLPDataset:
        return getattr(self, "_training_dataset", None)

    @property
    def validation_dataset(self) -> NLPDataset:
        return getattr(self, "_validation_dataset", None)

    def class2idx(self, class_name: str) -> Optional[int]:
        return self._class2idx.get(class_name, None)

    def idx2class(self, class_idx: int) -> Optional[str]:
        return self._idx2class.get(class_idx, None)

    def get_inputs(self) -> tf.Tensor:
        if getattr(self, "inputs", None) is None:
            return self.text2vec.get_inputs()
        else:
            return self.inputs

    def get_outputs(self) -> tf.Tensor:
        return self.build_output_layer(self.build_encode_layer(self.get_inputs()))

    def build(self) -> None:
        if self._built:
            return
        assert self._text2vec is not None
        super().build()

    def create_dataset_from_csv(
        self, filename: str, no_label: bool = False
    ) -> NLPDataset:
        return self.dataset_class(
            self.text2vec.tokenize,
            self.classes,
            csv_file=filename,
            max_length=self.max_sequence_length,
            no_label=no_label,
            **self.dataset_kwargs
        )

    def prepare_dataset(
        self,
        X: Sequence[Any],
        y: Sequence[Any],
        dataset: NLPDataset,
    ) -> NLPDataset:
        assert X is not None or dataset is not None
        if dataset is not None:
            return dataset
        return self.dataset_class(
            self.text2vec.tokenize,
            self.classes,
            X=X,
            y=y,
            max_length=self.max_sequence_length,
            no_label=y is None,
            **self.dataset_kwargs
        )

    def compile_optimizer(self, optimizer_name, **kwargs) -> None:
        if not self._built:
            self.build()
        weight_decay = kwargs.pop("weight_decay", 0)
        if optimizer_name != "adam":
            optimizer = tf.keras.optimizers.deserialize(
                {"class_name": optimizer_name, "config": kwargs}
            )
        else:
            optimizer = tfa.optimizers.AdamW(weight_decay, **kwargs)
        self._model.compile(
            optimizer=optimizer,
            loss=self.get_loss(),
            metrics=self.get_metrics(),
        )

    def fit(
        self,
        X: Sequence[Any] = None,
        y: Sequence[Any] = None,
        *,
        dataset: NLPDataset = None,
        validation_X: Optional[Sequence[Any]] = None,
        validation_y: Optional[Sequence[Any]] = None,
        validation_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 10,
        optimizer: str = "adam",
        optimizer_kwargs: Optional[dict[str, Any]] = None,
        learning_rate: float = 1e-3,
        weight_decay: float = 0.0,
        clip: Optional[float] = 5.0,
        learning_rate_update_factor: float = 0.5,
        learning_rate_update_epochs: int = 10,
        learning_rate_warmup_steps: int = 0,
        enable_early_stopping: bool = False,
        early_stopping_patience: Optional[int] = None,
        early_stopping_min_delta: float = 0.0,
        early_stopping_use_best_epoch: bool = False,
        early_stopping_monitor: int = 2,  # 1 for loss, 2 for metric
        checkpoint: Optional[str] = None,
        log_file: Optional[str] = None,
        verbose: int = 2
    ) -> None:
        if self.text2vec is None:
            raise ValueError("训练前必须先设置Text2vec")

        self._training_dataset = self.prepare_dataset(X, y, dataset)
        if (
            validation_X is None or validation_y is None
        ) and validation_dataset is None:
            self._validation_dataset = None
        else:
            self._validation_dataset = self.prepare_dataset(
                validation_X, validation_y, validation_dataset
            )
        training_tf_dataset = self.training_dataset.batchify(batch_size)
        validation_tf_dataset = None
        if self.validation_dataset is not None:
            validation_tf_dataset = self.validation_dataset.batchify(
                batch_size, shuffle=False
            )

        optimizer_kwargs = optimizer_kwargs or dict()
        optimizer_kwargs = {
            "learning_rate": learning_rate,
            "weight_decay": weight_decay,
            **optimizer_kwargs,
        }
        if clip is not None:
            optimizer_kwargs["clipnorm"] = clip
        self.compile_optimizer(optimizer, **optimizer_kwargs)

        monitor = "val_loss"
        monitor_direction = "min"
        if early_stopping_monitor == 2 and self.get_monitor():
            monitor = self.get_monitor()
            monitor_direction = "max"

        has_validation_dataset = validation_tf_dataset is not None
        if checkpoint is not None:
            if has_validation_dataset:
                checkpoint = os.path.join(checkpoint, self._get_model_filename(epoch=0))
            else:
                checkpoint = os.path.join(
                    checkpoint, self._get_model_filename_template()
                )
        callbacks = default_supervised_model_callbacks(
            learning_rate_update_factor=learning_rate_update_factor,
            learning_rate_update_epochs=learning_rate_update_epochs,
            learning_rate_warmup_steps=learning_rate_warmup_steps,
            use_weight_decay=weight_decay > 0,
            has_validation_dataset=has_validation_dataset,
            enable_early_stopping=enable_early_stopping,
            early_stopping_monitor=monitor,
            early_stopping_monitor_direction=monitor_direction,
            early_stopping_patience=early_stopping_patience,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_use_best_epoch=early_stopping_use_best_epoch,
            checkpoint=checkpoint,
            log_file=log_file,
        )
        for callback in self.get_callbacks():
            callback.validation_data = validation_tf_dataset
            callbacks.append(callback)

        self._model.fit(
            training_tf_dataset,
            epochs=n_epochs,
            validation_data=validation_tf_dataset if self.validation_dataset else None,
            callbacks=callbacks,
            verbose=verbose,
        )

    def predict(
        self,
        X: Sequence[Any] = None,
        *,
        dataset: NLPDataset = None,
        thresholds: Union[float, list[float], None] = None,
        batch_size: int = 128
    ) -> Union[np.ndarray, Sequence[Any]]:
        assert self._built
        dataset = self.prepare_dataset(X, None, dataset)
        return self._model.predict(dataset.batchify(batch_size, shuffle=False))

    def score(
        self,
        X: Sequence[Any] = None,
        y: Sequence[Any] = None,
        *,
        dataset: NLPDataset = None,
        thresholds: Union[float, list[float], None] = None,
        batch_size: int = 128
    ) -> pd.DataFrame:
        raise NotImplementedError()

    @classmethod
    def format_score(self, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "classes": self.classes,
            "task": self._task,
            "algorithm": self._algorithm,
        }

    def save(self, directory: str) -> None:
        super().save(directory)
        self.text2vec.save_vocab(directory)

    @classmethod
    def load(cls, directory: str, epoch: Optional[int] = None) -> "SupervisedNLPModel":
        module = super().load(directory, epoch=epoch)
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        module.text2vec = Text2vec(
            vocab,
            segmenter=module.segmenter,
            max_sequence_length=module.max_sequence_length,
            sequence_length=module.sequence_length,
        )
        return module

    def export(self, directory: str, name: str, version: str = "0") -> None:
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self.text2vec.save_vocab(d)


from __future__ import annotations
from typing import Sequence, Any, Optional, Callable

import tensorflow as tf
from tensorflow.keras.initializers import TruncatedNormal
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from official.nlp.keras_nlp import layers
import tensorflow_text as tftext

from sknlp.activations import gelu

from .bert_tokenization import BertTokenizationLayer


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BertCharPreprocessingLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        vocab: Sequence[str],
        cls_token: str = "[CLS]",
        sep_token: str = "[SEP]",
        **kwargs
    ) -> None:
        vocab: list = list(vocab)
        self.cls_token = cls_token
        self.sep_token = sep_token
        if cls_token not in vocab:
            vocab.append(cls_token)
        if sep_token not in vocab:
            vocab.append(sep_token)
        if "[UNK]" in vocab:
            idx = vocab.index("[UNK]")
            vocab[idx] = "unk"
        self.vocab = vocab
        self.tokenizer = TextVectorization(
            max_tokens=len(vocab),
            standardize=None,
            split=BertTokenizationLayer(cls_token=cls_token, sep_token=sep_token),
        )
        self.tokenizer.set_vocabulary(vocab[2:])
        super().__init__(**kwargs)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self.tokenizer(inputs)

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "cls_token": self.cls_token,
            "sep_token": self.sep_token,
        }


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BertPreprocessingLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        vocab: Sequence[str],
        cls_token: str = "[CLS]",
        sep_token: str = "[SEP]",
        **kwargs
    ) -> None:
        vocab: list = list(vocab)
        self.cls_token = cls_token
        self.sep_token = sep_token
        if cls_token not in vocab:
            vocab.append(cls_token)
        if sep_token not in vocab:
            vocab.append(sep_token)
        self.vocab = vocab
        for i, token in enumerate(vocab):
            if token == cls_token:
                self.cls_id = i
            if token == sep_token:
                self.sep_id = i
        super().__init__(**kwargs)

    def build(self, input_shape: tf.TensorShape) -> None:
        self.tokenizer = tftext.BertTokenizer(
            tf.lookup.StaticVocabularyTable(
                tf.lookup.KeyValueTensorInitializer(
                    self.vocab,
                    list(range(len(self.vocab))),
                    key_dtype=tf.string,
                    value_dtype=tf.int64,
                ),
                1,
            )
        )
        super().build(input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        token_ids = self.tokenizer.tokenize(inputs)
        flatten_ids = token_ids.merge_dims(-2, -1)
        cls_ids = tf.reshape(
            tf.tile(tf.constant([self.cls_id], dtype=tf.int64), [token_ids.nrows()]),
            [token_ids.nrows(), 1],
        )
        sep_ids = tf.reshape(
            tf.tile(tf.constant([self.sep_id], dtype=tf.int64), [token_ids.nrows()]),
            [token_ids.nrows(), 1],
        )
        ids = tf.concat([cls_ids, flatten_ids, sep_ids], axis=1)
        return ids.to_tensor(0)

    def get_config(self) -> dict[str, Any]:
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "cls_token": self.cls_token,
            "sep_token": self.sep_token,
        }


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BertPairPreprocessingLayer(BertPreprocessingLayer):
    def call(self, inputs: list[tf.Tensor]) -> list[tf.Tensor]:
        query, context = inputs
        query_token_ids = self.tokenizer.tokenize(query)
        context_token_ids = self.tokenizer.tokenize(context)
        query_flatten_ids = query_token_ids.merge_dims(-2, -1)
        context_flatten_ids = context_token_ids.merge_dims(-2, -1)
        cls_ids = tf.reshape(
            tf.tile(
                tf.constant([self.cls_id], dtype=tf.int64), [query_token_ids.nrows()]
            ),
            [query_token_ids.nrows(), 1],
        )
        sep_ids = tf.reshape(
            tf.tile(
                tf.constant([self.sep_id], dtype=tf.int64), [query_token_ids.nrows()]
            ),
            [query_token_ids.nrows(), 1],
        )
        query_ids = tf.concat([cls_ids, query_flatten_ids, sep_ids], axis=1)
        context_ids = tf.concat([context_flatten_ids, sep_ids], axis=1)
        type_ids = tf.concat(
            [
                tf.zeros_like(query_ids, dtype=tf.int64),
                tf.ones_like(context_ids, dtype=tf.int64),
            ],
            axis=1,
        )
        return (
            tf.concat([query_ids, context_ids], axis=1, name="token_ids").to_tensor(0),
            type_ids.to_tensor(0),
        )


@tf.keras.utils.register_keras_serializable(package="sknlp")
class BertEncodeLayer(tf.keras.layers.Layer):
    """
    copy from https://github.com/tensorflow/models/blob/v2.2.0/official/nlp/modeling/networks/transformer_encoder.py.
    """

    def __init__(
        self,
        vocab_size: int = 100,
        embedding_size: Optional[int] = 768,
        hidden_size: int = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: int = None,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: Callable = gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        share_layer: bool = False,
        cls_pooling: bool = True,
        name: str = "bert_encode_layer",
        **kwargs
    ) -> None:
        if embedding_size is None:
            embedding_size = hidden_size

        self.vocab_size = vocab_size
        self.embedding_size = embedding_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.sequence_length = sequence_length
        self.max_sequence_length = max_sequence_length
        self.type_vocab_size = type_vocab_size
        self.intermediate_size = intermediate_size
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.attention_dropout_rate = attention_dropout_rate
        self.initializer = initializer
        self.share_layer = share_layer
        self.cls_pooling = cls_pooling
        self.supports_masking = True
        super().__init__(name=name, **kwargs)

    def build(self, input_shape: tf.TensorShape) -> None:
        input_shape = input_shape[0]
        self.embedding_layer = layers.OnDeviceEmbedding(
            self.vocab_size,
            self.embedding_size,
            initializer=self.initializer,
            name="word_embeddings",
        )
        # Always uses dynamic slicing for simplicity.
        self.position_embedding_layer = layers.PositionEmbedding(
            self.max_sequence_length,
            initializer=self.initializer,
            name="position_embedding",
        )
        self.type_embedding_layer = layers.OnDeviceEmbedding(
            self.type_vocab_size,
            self.embedding_size,
            initializer=self.initializer,
            use_one_hot=True,
            name="type_embeddings",
        )
        if self.embedding_size != self.hidden_size:
            self.embedding_projection = tf.keras.layers.experimental.EinsumDense(
                "...x,xy->...y",
                output_shape=self.hidden_size,
                bias_axes="y",
                kernel_initializer=self.initializer,
                name="embedding_projection",
            )
        else:
            self.embedding_projection = None
        self.normalize_layer = tf.keras.layers.LayerNormalization(
            name="embeddings/layer_norm", axis=-1, epsilon=1e-12, dtype=tf.float32
        )
        self.embedding_dropout_layer = tf.keras.layers.Dropout(
            rate=self.dropout_rate, name="embeddings/dropout"
        )
        if self.share_layer:
            self.shared_layer = layers.TransformerEncoderBlock(
                self.num_attention_heads,
                self.intermediate_size,
                self.activation,
                output_dropout=self.dropout_rate,
                attention_dropout=self.attention_dropout_rate,
                kernel_initializer=self.initializer,
                name="transformer",
            )
        else:
            self.transformer_layers = []
            for i in range(self.num_layers):
                layer = layers.TransformerEncoderBlock(
                    self.num_attention_heads,
                    self.intermediate_size,
                    self.activation,
                    output_dropout=self.dropout_rate,
                    attention_dropout=self.attention_dropout_rate,
                    kernel_initializer=self.initializer,
                    name="transformer/layer_%d" % i,
                )
                self.transformer_layers.append(layer)

        if self.cls_pooling:
            self.cls_output_layer = tf.keras.layers.Dense(
                units=self.hidden_size,
                activation="tanh",
                kernel_initializer=self.initializer,
                name="pooler_transform",
            )
        super().build(input_shape)

    def call(self, inputs: list[tf.Tensor]) -> list[tf.Tensor]:
        token_ids, type_ids = inputs
        mask = self.compute_mask(inputs, mask=None)[0]
        word_embeddings = self.embedding_layer(token_ids)
        position_embeddings = self.position_embedding_layer(word_embeddings)
        type_embeddings = self.type_embedding_layer(type_ids)
        embeddings = tf.keras.layers.Add()(
            [word_embeddings, position_embeddings, type_embeddings]
        )
        embeddings = self.normalize_layer(embeddings)
        embeddings = self.embedding_dropout_layer(embeddings)
        if self.embedding_projection is not None:
            embeddings = self.embedding_projection(embeddings)
        data = embeddings
        attention_mask = layers.SelfAttentionMask()(data, mask)
        encoder_outputs = []
        if self.share_layer:
            for _ in range(self.num_layers):
                data = self.shared_layer([data, attention_mask])
                encoder_outputs.append(data)
        else:
            for layer in self.transformer_layers:
                data = layer([data, attention_mask])
                encoder_outputs.append(data)

        first_token_tensor = tf.keras.layers.Lambda(
            lambda x: tf.squeeze(x[:, 0:1, :], axis=1)
        )(encoder_outputs[-1])
        cls_output = first_token_tensor
        if self.cls_pooling:
            cls_output = self.cls_output_layer(cls_output)
        return [encoder_outputs, cls_output]

    def compute_mask(
        self, inputs: list[tf.Tensor], mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        if mask is not None:
            return [mask for _ in range(self.num_layers)]
        return [tf.math.not_equal(inputs[0], 0) for _ in range(self.num_layers)]

    def get_config(self):
        return {
            **super().get_config(),
            "vocab_size": self.vocab_size,
            "embedding_size": self.embedding_size,
            "hidden_size": self.hidden_size,
            "num_layers": self.num_layers,
            "num_attention_heads": self.num_attention_heads,
            "sequence_length": self.sequence_length,
            "max_sequence_length": self.max_sequence_length,
            "type_vocab_size": self.type_vocab_size,
            "intermediate_size": self.intermediate_size,
            "activation": self.activation,
            "dropout_rate": self.dropout_rate,
            "attention_dropout_rate": self.attention_dropout_rate,
            "initializer": self.initializer,
            "share_layer": self.share_layer,
            "cls_pooling": self.cls_pooling,
        }


# -*- coding: utf-8 -*-
from .classification_dataset import ClassificationDataset
from .tagging_dataset import TaggingDataset
from .similarity_dataset import SimilarityDataset
from .nlp_dataset import NLPDataset
from .bert_dataset import (
    BertClassificationDataset,
    BertTaggingDataset,
    BertSimilarityDataset,
)


__all__ = [
    "NLPDataset",
    "ClassificationDataset",
    "TaggingDataset",
    "BertClassificationDataset",
    "BertTaggingDataset",
    "SimilarityDataset",
    "BertSimilarityDataset",
]
